"""
Basic building blocks for generic class based views.

We don't bind behaviour to http method handlers yet,
which allows mixin classes to be composed in interesting ways.
"""
from rest_framework import status
from rest_framework.response import Response
from rest_framework.settings import api_settings
from .utils import APIResponse, reset_db_sequence
from warnings import warn
from rest_framework.mixins import CreateModelMixin, UpdateModelMixin, DestroyModelMixin


class MyCreateModelMixin(CreateModelMixin):
    """
    Create a model instance.
    """
    def create(self, request, *args, **kwargs):
        data = request.data
        serializer = self.get_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        return APIResponse(serializer.data, status=201, msg='ok, 创建成功.')

    def perform_create(self, serializer):
        try:
            serializer.save()
        except Exception as e:
            warn('Warning: 序列化器保存错误! 可能是最近有csv/excel数据导入引起的主键冲突. \n详细信息:' + str(e))
            reset_db_sequence(self.queryset)
            serializer.save()


class MyUpdateModelMixin(UpdateModelMixin):
    def update(self, request, *args, **kwargs):
        partial = kwargs.pop('partial', False)
        instance = self.get_object()
        serializer = self.get_serializer(instance, data=request.data, partial=partial)
        serializer.is_valid(raise_exception=True)
        self.perform_update(serializer)

        if getattr(instance, '_prefetched_objects_cache', None):
            # If 'prefetch_related' has been applied to a queryset, we need to
            # forcibly invalidate the prefetch cache on the instance.
            instance._prefetched_objects_cache = {}

        return APIResponse(serializer.data, status=200, msg='ok, 更新成功.')


class MyDestroyModelMixin:
    """
    Destroy a model instance.
    """
    def destroy(self, request, *args, **kwargs):
        instance = self.get_object()
        # try:
        #     instance = self.get_object()
        # except Exception as e:
        #     return APIResponse(None, status=403, msg='Error! error_msg:'+ str(e))
        self.perform_destroy(instance)
        return APIResponse(None, status=status.HTTP_204_NO_CONTENT, msg='ok, 删除成功.')

    def perform_destroy(self, instance):
        instance.delete()




"""
解析introduction, 自动生成.md

- 示例:
```
标题
- 这一定是简介, 没有的话, 简介就是标题

GET /api/authors/InstitutionType/   # 接口, 可能会在最前方指明请求类型[GET, POST]
POST /api/authors/InstitutionType/      # 注明POST方法的话, 会将后面携带的dict加上
{
    "请求数据": "xxx"
}
```

"""


import os
import sys
from jinja2 import Template
from bdtime import tt

from django.conf import settings
from django.http import HttpResponse
from django.shortcuts import render
from django.core.cache import cache
from rest_framework.request import Request
from rest_framework.views import APIView

from .django import APIResponse, BaseListView, api_decorator
from .MultipleConditionSearch import AdvancedSearchView
from .django import convert_db_field_type_to_python_type
from .pure import convert_query_parameter_to_bool
from .django import get_base_serializer
from .django import my_api_assert_function


global tmp


RLT_PATH_OF_JINJIA2_TEMPLATE = 'templates\\autoWiki.html'       # 使用`jinja2`模板来生成`wiki`模板
ABS_PATH_OF_JINJIA2_TEMPLATE = os.path.join(os.path.dirname(__file__), RLT_PATH_OF_JINJIA2_TEMPLATE)


class AutoWiki(APIView):
    """
    自动生成wiki文档

    - app_name: 目标app, 必填
    - view_class_name: 目标view, 可空
    - ret_all_fields: 是否返回全部字段, 默认False
    - ADD_CAN_BE_EMPTY: 增加“能否为空”字段, 默认False
    - get_output_file: 下载文件, 仅能下一个*view_class*的说明文档, 适合*linux*服务端上使用

    - `.md`备注方法
        {#[*]( "这是注释")#}
        {#[ ]( "最好写上 [GET POST] 等")#}
    """

    open_output_txt_file = True     # `win`在运行完成后, 打开输出的`txt`文件
    path_of_jinja2_template = ABS_PATH_OF_JINJIA2_TEMPLATE       # 使用`jinja2`模板来生成`wiki`模板

    # response_render_format = {
    #     'status': 'code',
    #     'msg': 'message',
    #     'result': 'data',
    # }
    ret_dc = {
        'status': 200,
        'msg': 'ok',
        'result': {
            'page_dc': {
                'count_items': 10,
                'total_pages': 1,
                'page_size': 3,
                'p': 1,
            },
            'data': None,
        },
    }

    def get(self, request, *args, **kwargs):
        """自动生成wiki文档"""
        res_context_dc = {}     # 将结果变量储存为文本字典, 好用`jinja2`来填写模板

        ret = []
        request_data = request.GET

        # --- 必填
        app_name = request_data.get('app_name')
        # assert app_name, 'app_name不能为空!'
        my_api_assert_function(app_name, 'app_name不能为空!')

        # --- 选填
        view_class_name = request_data.getlist('view_class_name')       # 可空
        ret_all_fields = request_data.getlist('ret_all_fields', False)     # 返回全部字段 or 只保留serializer_fields中返回的字段
        ADD_CAN_BE_EMPTY = request_data.getlist('ADD_CAN_BE_EMPTY', False)   # 增加“能否为空”字段
        SLEEP_TIME = request_data.get('SLEEP_TIME', 0.2)    # sleep时间, 也就是文件保存到打开的间隔
        MAX_LEN = request_data.get('MAX_LEN', 100)      # 示例数据最长字段限制

        exec(f'global tmp; from {app_name} import views as tmp;')        # 执行
        print('---------', tmp)

        views = tmp
        func_ls = dir(views)

        s = 0
        vs = []
        for f in func_ls:
            md = getattr(views, f)
            if hasattr(md, 'queryset') and (hasattr(md, 'serializer_class') or hasattr(md, 'list_fields') or hasattr(md, 'retrieve_fields')):
                print('---------', f, md)
                if getattr(md, 'queryset') is not None and (getattr(md, 'serializer_class') or getattr(md, 'list_serializer_class') or getattr(md, 'auto_generate_serializer_class') or getattr(md, 'retrieve_serializer_class')) or getattr(md, 'list_fields') or getattr(md, 'retrieve_fields'):
                    if not view_class_name or f in view_class_name:
                        s += 1
                        print(f, md)
                        vs.append([f, md])
        print(s)
        if s == 0:
            msg = '没有任何可以转换的view_class! 将自动生成!!'
            return APIResponse(None, status=404, msg=msg)

        print('--------------------------------------------------------------------------')

        """生成一个view的wiki文档"""

        """生成一个view的wiki文档"""
        from bddjango.django import get_base_model
        from django.forms import model_to_dict
        from bddjango import show_json
        from bddjango.django import get_field_type_in_py
        from bddjango import create_dir_if_not_exist, create_file_if_not_exist
        import re
        import json
        import os

        tempdir_rootpath = 'tempdir'        # 临时输出文件的根目录
        output_dirpath = os.path.join(tempdir_rootpath, 'output')       # 临时输出文件的子目录
        create_dir_if_not_exist(tempdir_rootpath)
        create_dir_if_not_exist(output_dirpath)

        path_of_jinja2_template = request_data.get('path_of_jinja2_template', self.path_of_jinja2_template)

        # region # --- 循环获取view的文档

        for view_id in range(len(vs)):
            # view_id = 0       # debug
            # issubclass(v, BaseListView)

            # --- 获取单个view的名称和实例
            fv = vs[view_id]
            f, v = fv       # view_class_name, view_class_instance

            # --- 输出文件的路径
            output_f_suffix = '.txt'        # 使用`.md`打开太慢, `typora`会报错
            # output_f_suffix = '.md' if path_of_jinja2_template else '.txt'

            output_fname = f + output_f_suffix
            output_fname = os.path.join(output_dirpath, output_fname)
            if os.path.exists(output_fname):
                os.remove(output_fname)

            def output_to_file_and_prt(text='', output_fname=output_fname, show_in_console=False):

                output_file = open(output_fname, 'a+', encoding='utf-8')

                if show_in_console:
                    print(text)
                print(text, file=output_file)
                output_file.close()

            # --- 开始写入

            output_to_file_and_prt('****************')
            res_context_dc['introduction'] = v.__doc__      # context: 简介
            output_to_file_and_prt(res_context_dc['introduction'])

            output_to_file_and_prt('****************')
            output_to_file_and_prt()

            output_to_file_and_prt(f'\n=======  {view_id}  =========  \n')
            res_context_dc['view_id'] = str(view_id)  # context: view_id

            view_class_name, view_class_instance = f, v
            output_to_file_and_prt(f'---, view_class_name: [{view_class_name}], view_class_instance: [{view_class_instance}]')
            res_context_dc['view_class_name'] = str(view_class_name)     # context: view_class_name
            res_context_dc['view_class_instance'] = str(view_class_instance)     # context: view_class_instance

            output_to_file_and_prt()

            # --- 获取model和meta
            md = get_base_model(v.queryset)
            meta = md._meta

            url = f'/api/{app_name}/{f}/'

            output_to_file_and_prt('------------ 请求URL')
            output_to_file_and_prt(f"`{url}`")
            res_context_dc['request_url'] = str(url)     # context: 请求URL

            output_to_file_and_prt()

            filter_fields = None
            if hasattr(v, 'filter_fields'):
                if v.filter_fields:
                    output_to_file_and_prt('------------ 过滤字段')
                    output_to_file_and_prt(f"`{v.filter_fields}`")

                    filter_fields = v.filter_fields
                    if filter_fields == []:
                        filter_fields = None

                    res_context_dc['filter_fields'] = str(v.filter_fields)  # context: 请求URL
                    output_to_file_and_prt()

            # 确定serializer_class, 以免报错
            v.serializer_class = v.serializer_class or v.retrieve_serializer_class or v.list_serializer_class

            # --- 获取field_name和verbose_name

            field_names = [field.name for field in meta.fields]
            verbose_names = [field.verbose_name for field in meta.fields]
            can_be_empty_ls = [field.null and field.blank for field in meta.fields]

            field_type_ls = [get_field_type_in_py(md, field_name) for field_name in field_names]
            # field_type = get_field_type_in_py(md, field_name)

            if hasattr(v, 'auto_generate_serializer_class') and v.auto_generate_serializer_class:
                # from bddjango import get_base_serializer
                # v: BaseListView
                v__serializer_class = v().get_serializer_class()
                # v__queryset = get_base_model(v().get_queryset()).objects.all()
                # v__serializer_class(v__queryset, many=True).data
                # v__serializer_class.__dict__.get('_declared_fields')
                v.serializer_class = v__serializer_class

            if not v.serializer_class:
                list_serializer_class = retrieve_serializer_class = None
                if hasattr(v, 'list_fields'):
                    list_fields = getattr(v, 'list_fields')
                    list_serializer_class = get_base_serializer(v.queryset, list_fields)

                if hasattr(v, 'retrieve_fields'):
                    retrieve_fields = getattr(v, 'retrieve_fields')
                    retrieve_serializer_class = get_base_serializer(v.queryset, retrieve_fields)
                v.serializer_class = list_serializer_class or retrieve_serializer_class

            # --- 把serializers里面的拓展字段通过__doc__转义后加进field_names
            serializer_field_ls = v.serializer_class.__dict__.get('_declared_fields')
            text = v.serializer_class.__doc__ or ""

            for serializer_field_i, field_type_i in serializer_field_ls.items():
                print(serializer_field_i, field_type_i)
                reg = re.compile(f'^.*{serializer_field_i}: +(.*?) *$', re.M)
                match = reg.search(text)
                if match:
                    verbose_name_i = match.group(1)
                else:
                    verbose_name_i = serializer_field_i

                _conved_flag = False
                if isinstance(serializer_field_i, str) and '__' in serializer_field_i:
                    fk_model_name, fk_field_name = serializer_field_i.split('__', 1)
                    # print('--- 序列化自动生成的字段: ', fk_model_name, fk_field_name)
                    if fk_model_name in field_names and '__' not in fk_field_name:
                        _meta_field = meta.fields[field_names.index(fk_model_name)]
                        # to_field = _meta_field.to_fields[0]
                        # to_field = to_field if to_field else 'pk'
                        _field_value = getattr(md, fk_model_name)
                        from .django.utils import get_field_names_by_model
                        from .django.utils import get_field_type_in_py

                        related_fields = get_field_names_by_model(_meta_field.related_model)
                        _fields = _meta_field.related_model._meta.fields
                        _real_fk_field_name = _fields[related_fields.index(fk_field_name)].name
                        field_type_i = get_field_type_in_py(_meta_field.related_model, _real_fk_field_name)
                        _conved_flag = True
                    else:
                        if '__' in fk_field_name:
                            print(f'*** 警告: 仅支持一层外键自动化转换类型! 自动转换失败字段: `{serializer_field_i}`')

                if not _conved_flag:
                    field_type_i = convert_db_field_type_to_python_type(str(field_type_i).replace('()', ''))
                print(serializer_field_i, verbose_name_i, field_type_i)
                can_be_empty_i = True

                field_names.append(serializer_field_i)
                verbose_names.append(verbose_name_i)
                field_type_ls.append(field_type_i)
                can_be_empty_ls.append(can_be_empty_i)

            # --- 获取serializer_class
            serializer_class = v.serializer_class

            if issubclass(v, BaseListView):
                if serializer_class is None:
                    if v.list_serializer_class:
                        serializer_class = v.list_serializer_class
                    if v.retrieve_serializer_class:
                        # 有限详情页的序列化器
                        serializer_class = v.retrieve_serializer_class

            # --- 示例字段
            smeta = serializer_class.Meta
            if hasattr(smeta, 'fields'):
                sf = serializer_class.Meta.fields
                if ret_all_fields is True or sf == '__all__':
                    # sf = '__all__'
                    sf = serializer_class.Meta.fields = field_names
            else:
                sf = field_names.copy()
                exclude_ls = smeta.exclude
                if exclude_ls:
                    for e in exclude_ls:
                        sf.remove(e)

            try:
                # 直接调用view的get接口
                result = v.as_view()(request._request).data.get('result')

                if isinstance(result, dict):
                    dc_ls = result.get('data')
                elif isinstance(result, list):
                    dc_ls = result
                else:
                    raise TypeError(f'result为未知的返回类型! {type(result)}')
                dc_ls = dc_ls[:3]
                if not dc_ls:
                    raise ValueError('\n============== 返回的样例数据`example_data`为空! =============\n')

            except Exception as e:
                try:
                    print('--- Error! 调用view方法失败, 尝试使用model方案... ', e)
                    # --- 示例数据, 用model
                    qs_ls = md.objects.all()[:3]
                    # q0 = qs_ls[0]     # 单个数据
                    try:
                        dc_ls = serializer_class(qs_ls, many=True).data
                    except Exception as e:
                        print('--- Error! 序列化出错! 可能未携带annotate所需字段? 将自动生成基本序列化器! --- ' + str(e))
                        serializer_class = get_base_serializer(qs_ls)
                        dc_ls = serializer_class(qs_ls, many=True).data

                    dc_ls = [dict(dc) for dc in dc_ls]      # 返回数据示例
                except Exception as e:
                    print(f'========== 错误信息: view_class_name[{view_class_name}], view_class_instance[{v}]')
                    raise e

            # -- 简化过长的字段
            for dc in dc_ls:
                for k, v in dc.items():
                    if isinstance(v, str):
                        if len(v) > MAX_LEN:
                            dc[k] = v[:MAX_LEN] + '...略'
                print(dc)

            output_to_file_and_prt('---------- 示例数据')

            _ret_dc = self.ret_dc.copy()
            _ret_dc['result']['data'] = dc_ls

            # 整理`response`的格式
            response_render_format = {
                'status': 'status',
                'msg': 'msg',
                'result': 'result',
            }
            if hasattr(self, 'response_render_format') and getattr(self, 'response_render_format'):
                response_render_format = self.response_render_format
            res_context_dc['response_render_format'] = response_render_format

            for k, v in response_render_format.items():
                _ret_dc[v] = _ret_dc.pop(k)
            example_data = json.dumps(_ret_dc, sort_keys=False, indent=4, separators=(', ', ': '), ensure_ascii=False)

            output_to_file_and_prt(example_data)
            res_context_dc['example_data'] = str(example_data)  # context: 示例数据
            output_to_file_and_prt()

            # --- 参数说明
            print(field_names, verbose_names)

            # region # --- 分析属于那种请求类型`context_request_type`
            view_class_type = 'APIView'
            if hasattr(view_class_instance, '_name'):
                view_class_type = getattr(view_class_instance, '_name')

            post_type_ls = getattr(view_class_instance, 'post_type_ls') if hasattr(view_class_instance,
                                                                                   'post_type_ls') else None

            context_request_type = '其它'
            if view_class_type == 'BaseListView':
                context_request_type = '基本查找类'
            elif view_class_type == 'CompleteModelView':
                context_request_type = '增删查改类'
            elif view_class_type == 'AdvancedSearchView':
                context_request_type = '高级检索类'
            elif view_class_type == 'BaseFullTextSearchView':
                context_request_type = '全文检索类'
            res_context_dc['context_request_type'] = context_request_type
            res_context_dc['view_class_type'] = view_class_type
            res_context_dc['post_type_ls'] = post_type_ls
            # endregion

            filter_fields__doc = ""
            res_context_dc['has_filter_fields'] = False
            if filter_fields:
                filter_fields__doc = "| 类型 | 字段名 | 说明 | 必填 |\n| --- | --- | --- | --- |\n"
                res_context_dc['has_filter_fields'] = True

            if view_class_type == 'BaseFullTextSearchView':
                filter_fields__doc = "| 类型 | 字段名 | 说明 | 必填 |\n| --- | --- | --- | --- |\n"
                filter_fields__doc += "| str  | search_keywords   | 全文检索关键词   | 否 |\n"
                _p = 'use_relevant_id_param'
                if hasattr(view_class_instance, _p) and getattr(view_class_instance, _p):
                    filter_fields__doc += "| int | relevant_id   | 将要做相关推荐的对象id     | 否 |\n"

            if ADD_CAN_BE_EMPTY:
                ss = "| 类型 | 字段名 | 说明 | 必填 |\n| --- | --- | --- | --- |\n"
            else:
                ss = "| 类型 | 字段名 | 说明 |\n| --- | --- | --- |\n"

            for field_name, verbose_name, can_be_empty, field_type in zip(field_names, verbose_names, can_be_empty_ls, field_type_ls):
                print(field_name, verbose_name, can_be_empty)

                si = ''
                if sf == '__all__' or field_name in sf:
                    # field_type = get_field_type_in_py(md, field_name)
                    if ADD_CAN_BE_EMPTY:
                        must_fill = '否' if can_be_empty else '是'
                        si = f'| {field_type} | {field_name} | {verbose_name} | { must_fill } |\n'
                    else:
                        si = f'| {field_type} | {field_name} | {verbose_name} |\n'
                ss += si

                if filter_fields and (filter_fields == '__all__' or field_name in filter_fields):
                    filter_fields__doc += f'| {field_type} | {field_name} | {verbose_name} | 否 |\n'

            output_to_file_and_prt('---------- 参数说明')
            parameters_explain = ss
            output_to_file_and_prt(parameters_explain)
            res_context_dc['parameters_explain'] = str(parameters_explain)  # context: 参数说明

            res_context_dc['filter_fields__doc'] = str(filter_fields__doc)  # filter_fields__doc: 过滤字段

            output_to_file_and_prt()

            print(parameters_explain)

            # if convert_query_parameter_to_bool(open_output_txt_file):
            if convert_query_parameter_to_bool(path_of_jinja2_template):
                # abs_path = os.path.join(settings.BASE_DIR, 'authors/myTestTemplates.html')
                # path_of_jinja2_template = path_of_jinja2_template if os.path.exists(path_of_jinja2_template) else os.path.exists(os.path.join(settings.BASE_DIR, path_of_jinja2_template))
                path_of_jinja2_template = os.path.join(settings.BASE_DIR, path_of_jinja2_template)
                if sys.platform == 'win32':
                    path_of_jinja2_template = path_of_jinja2_template.replace('/', '\\')
                else:
                    path_of_jinja2_template = path_of_jinja2_template.replace('\\', '/')

                my_api_assert_function(os.path.exists(path_of_jinja2_template), f'path_of_jinja2_template路径不存在![{path_of_jinja2_template}]')
                print('--- res_context_dc:', res_context_dc)

                # region # --- 解析`introduction`, 获取[标题, 简要描述, 请求url]等字段
                introduction = res_context_dc.get('introduction')
                if introduction:
                    # --- 获取标题`introduction_title`
                    reg = re.compile(r'^\n(.+?)\n')
                    match = reg.search(introduction)
                    if match:
                        introduction_title = match.group(1).strip()
                        introduction_title = re.sub(r'^#+ +', '', introduction_title)   # 去掉开头的`#`
                        res_context_dc['introduction_title'] = introduction_title

                    # --- 获取示例url`introduction_url`
                    reg = re.compile(r'.*?/.+/.+/.*')
                    match = reg.findall(introduction)
                    if match:
                        _introduction_url = []
                        for m_i in match:
                            m_i = m_i.strip()
                            if 'POST' in m_i:
                                reg = re.compile(m_i + '.*?' + r'\n.*(\{.*?\})[ \t]*\n', re.S)      # 只能匹配POST后的一个大括号!
                                # m = reg.search(introduction)
                                # if m:
                                #     text = m.group(1)
                                #     reg = re.compile('^.*(\{.*?\})\n', re.S)
                                #     _m = reg.search(text)
                                #     _m and print(_m.group(1))
                                #     dc = json.dumps(json.loads(_m.group(1)), sort_keys=False, indent=4, separators=(', ', ': '), ensure_ascii=False)
                                findall_ls = reg.findall(introduction)
                                # print(findall_ls)
                                if findall_ls:
                                    findall_i = findall_ls[0]
                                    try:
                                        findall_str = json.dumps(json.loads(findall_i), sort_keys=False, indent=4, separators=(', ', ': '), ensure_ascii=False)
                                    except Exception as e:
                                        e_str = '****** POST请求携带的字典错误!' + str(e)
                                        from warnings import warn
                                        warn(e_str)
                                        # raise ValueError(e_str)
                                    m_i += '\n' + findall_str
                            _introduction_url.append(m_i)
                        introduction_url = '\n'.join(_introduction_url)
                        res_context_dc['introduction_url'] = introduction_url

                    reg = re.compile(r'\n( *?- .+)')
                    match = reg.findall(introduction)
                    # from bddjango import show_ls
                    # show_ls(match)
                    if match:
                        _introduction_summary = []
                        blank_ls = []
                        for summary_i in match:
                            summary_i: str
                            m1 = re.match(r'^ *', summary_i)
                            blank_i = m1.span()[-1] if m1 else 0
                            blank_ls.append(blank_i)
                            _introduction_summary.append(summary_i)
                        blank_min = min(blank_ls)       # 将所有简介缩进到最左侧, 不然会变成代码块!
                        _introduction_summary = [s_i[blank_min:] for s_i in _introduction_summary]
                        introduction_summary = '\n'.join(_introduction_summary)
                    else:
                        introduction_summary = '- ' + introduction_title
                    res_context_dc['introduction_summary'] = introduction_summary
                # endregion

                # region # --- 开始填充jinja模板
                with open(path_of_jinja2_template, encoding='utf-8') as file_:
                    template = Template(file_.read())

                # from jinja2 import PackageLoader, Environment, FileSystemLoader
                # env = Environment(loader=PackageLoader('python_project', 'templates'))  # 创建一个包加载器对象

                # env = Environment(loader=FileSystemLoader(app_name))  # 文件加载器, 可用`list_templates`方法查看存在哪些东西
                # os.path.exists(path_of_jinja2_template)
                # template = env.get_template(path_of_jinja2_template)
                content = template.render(**res_context_dc)
                with open(output_fname, 'w', encoding='utf-8') as f:
                    f.write(content)
                # endregion

            # 获取output文件, 仅能获取一个view_class
            get_output_file = convert_query_parameter_to_bool(request_data.get('get_output_file'))
            if get_output_file:
                newFileName = output_fname
                with open(output_fname, 'rb') as f:
                    response = HttpResponse(f.read(), content_type='application/txt')
                    response['Content-Disposition'] = 'attachment; filename="{}"'.format(newFileName)
                    response["Access-Control-Allow-Origin"] = '*'
                    response["Server"] = '*'
                    return response
            else:
                open_output_txt_file = request_data.get('open_output_txt_file', self.open_output_txt_file)
                if sys.platform in ['win32', 'darwin'] and convert_query_parameter_to_bool(open_output_txt_file):
                    for i in range(5):
                        tt.sleep(1)
                        print('111')
                        if os.path.exists(output_fname):
                            print('222')
                            if sys.platform == 'win32':
                                os.startfile(output_fname)
                            elif sys.platform == 'darwin':
                                os.system(f'open {output_fname}')
                            else:
                                my_api_assert_function(f'系统类型错误? sys.platform: {sys.platform}')
                            tt.sleep(SLEEP_TIME)
                            # tt.sleep(10)
                            # os.remove(output_fname)
                            from bddjango.adminclass import remove_temp_file
                            remove_temp_file('tempdir', MAX_TEMPS=10)
                            break
                else:
                    print('--- output_fname: ', output_fname)
                    with open(output_fname, 'r', encoding='utf-8') as f:
                        # print(f.read())

                        res_context_dc.update(
                            {
                                'output_fname': output_fname,
                                'output_content': f.read(),
                            }
                        )
            ret_dc = {
                'res_context_dc': res_context_dc
            }
            ret.append(ret_dc)

        # endregion

        # output_file.close()
        return APIResponse(ret, 200, 'ok')


from warnings import warn
from .pure import *


def version():
    """
    * 2021/6/1
    - [pypi_url](https://pypi.org/project/bddjango/)

    # 2.2
        - 方法version转移到__init__.py中
        - 修改由`get_key_from_query_dc_or_self`引起的bug
        - 高级检索类增加`search_ls_dc`和`search_conf`两个参数
        - 修改由于readme.md出错导致pypi上传失败的bug	# 2.2.2
        - 去掉部分print	# 2.2.3
        - 新增conv_to_queryset方法	# 2.2.3
        - 修复部分bug	# 2.2.4
    # 2.3 baseDjango项目
        - StateMsgResultJSONRenderer错误提示    # 2.3.0
        - autoWiki部分修改
        - _remove_temp_file修改为默认保存最近访问过的1/3缓存文件
        - BaseListView支持orm参数, 并可使用`convert_to_bool_flag`强制转换为bool变量
        - 拆分.django文件
        - 增加`.django.judge_db_is_migrating`, 来判断是否正在迁移数据库
        - 将`remove_temp_file`迁移至.pure中      # 2.3.1
        - 将adminclass中的导入导出类增加权限验证和自定义取消功能
        - 将adminclass中的导出action增加取消功能       # 2.3.2
    # 2.4 贵州图书馆项目
        - `get_base_serializer`增加`auto_generate_annotate_fields`功能
        - 修复了`run_list_filter`中的`annotate`字段检索失效的问题
        - `AutoWiki`增加mac兼容
        - BaseListView的retrieve修复queryset中annotate字段失效问题
        - 修复get_base_serializer在queryset进行values后annotate字段失效问题
        - 修复get_key_from_query_dc_or_self当query_dc中有False时返回值出错的bug
        - 修复get_key_from_query_dc_or_self当query_dc中获取bool错误问题       # 2.4.1
        - get_MySubQuery增加注释
        - order_qs_ls_by_id完善为不限制长度
        - zip功能完善       # 2.4.2
        - get_base_serializer解决'__all__'和retrieve时出现的bug        # 2.4.3
        - retrieve_filter_field现在可由前端指定
        - excel导入datetime字段时的处理
        - BaseListView中的count替换为exists, 提升性能
        - 修复exists的值取反导致检索全部失效的bug
        - 修复extract_pdf出错的bug: PdfFileReader(pdfFile, strict=False)
        - 返回页码p不能为0的报错信息
        - .adminclass增加注释和使用说明        # 2.4.4
        - bulk_delete增加权限控制        # 2.4.5
        - order_by_order_type_ls修复为[""]的时候出现的bug
        - 页码p的提示更人性化, 改为必须正整数        # 2.4.6
        - 页码p或page_size等于GET_ALL(-999)时, 则返回全部数据
        - admin增加orm检索      # 2.4.7
        - simplrui的admin_class的基类改为`AjaxAdmin`
        - 修复Mixin后置导致的Guadian失效的bug     # 2.4.8
        - 增加_ajax_return_qs_ls, 以便admin中的ajax方法返回qs_ls      # 2.4.9
        - django.utils.get_field_names_by_model增加`field_attr`, 以便获取[name, verbose_name]
        - BaseListView再次简化, 省略get_base_serializer步骤, 只需要list_fields和retrieve_fields即可       # 2.5.0
        - `.django.utils.get_field_names_by_model`增加`exclude_ls`功能, 并按原模型字段顺序排序
        - 完善高级检索功能, AddQS现在支持无限Q_ls的嵌套了     # 2.5.1
        - 过滤条件fn以`__isnull`结尾时, value自动转换为bool类型        # 2.6.0
        - autoWiki兼容list_fields和retrieve_fields
        - adminclass.MyAjaxAdmin导致和simpleui耦合              # 2.6.1
        - admin导入导出功能增加对外键的支持                       # 2.6.2
        - autoWiki模板优化
        - 清除admin导入功能的debug信息                       # 2.6.3
        - adminclass用.filter时增加Q函数支持
        - `get_base_serializer`增加对多层外键字段的支持
        - 解决autoWiki的模板bug(view_class_type引起的请求方式不对)                       # 2.6.4
        - autoWiki优化, 增加`BaseFullTextSearchView`类
        - 增加`baseSearchView.py`, 包含中文全文检索和相关对象推荐
        - 导入时保持id值的功能       # 2.6.5
        - 相关推荐`RelevantObjectRecommendationMixin`增加`relevant_base_model`字段
        - `renderer_classes`还原为APIView原始值, 避免浏览器templates功能失效
    """
    v = "2.6.5"  # 当前发布: 2.6.5
    return v


try:
    from .django import *
except Exception as e:
    warn('导入django失败? --- ' + str(e))

try:
    from .myFIelds import AliasField  # 这个只能在这里引用, 不然`adminclass`报错
except Exception as e:
    warn('导入`AliasField`失败? --- ' + str(e))



from django.utils.encoding import force_str
from rest_framework.utils.serializer_helpers import ReturnList, ReturnDict
from rest_framework.exceptions import APIException, ErrorDetail, status
from django.utils.translation import ugettext_lazy
from rest_framework.authentication import BaseAuthentication
from rest_framework import exceptions
from rest_framework.authtoken.models import Token
from rest_framework import HTTP_HEADER_ENCODING
from django.contrib.auth.models import AnonymousUser


def _get_error_details(data, default_code=None):
    """
    Descend into a nested data structure, forcing any
    lazy translation strings or strings into `ErrorDetail`.
    """
    if isinstance(data, list):
        ret = [
            _get_error_details(item, default_code) for item in data
        ]
        if isinstance(data, ReturnList):
            return ReturnList(ret, serializer=data.serializer)
        return ret
    elif isinstance(data, dict):
        ret = {
            key: _get_error_details(value, default_code)
            for key, value in data.items()
        }
        if isinstance(data, ReturnDict):
            return ReturnDict(ret, serializer=data.serializer)
        return ret

    text = force_str(data)
    code = getattr(data, 'code', default_code)
    return ErrorDetail(text, code)


class MyApiError(APIException):
    status_code = status.HTTP_200_OK
    default_detail = ('Invalid input.', )
    default_code = 'invalid'

    def __init__(self, detail=None, code=None):
        if detail is None:
            detail = self.default_detail
        if code is None:
            code = self.default_code

        # For validation failures, we may collect many errors together,
        # so the details should always be coerced to a list if not already.
        if not isinstance(detail, dict) and not isinstance(detail, list):
            detail = [detail]

        self.detail = _get_error_details(detail, code)


class MyValidationError(APIException):
    status_code = status.HTTP_200_OK
    default_detail = ('Invalid input.')
    default_code = 'invalid'

    def __init__(self, detail=None, code=None):
        if detail is None:
            detail = self.default_detail
        if code is None:
            code = self.default_code

        # For validation failures, we may collect many errors together,
        # so the details should always be coerced to a list if not already.
        if not isinstance(detail, dict) and not isinstance(detail, list):
            detail = [detail]

        self.detail = _get_error_details(detail, code)


def get_my_api_error(status=status.HTTP_404_NOT_FOUND, msg="Error!"):
    ret = MyApiError(detail={
                'status': status,
                'msg': msg,
                'result': [],
            })

    return ret


def my_api_assert_function(assert_sentence, msg='error', status=404):
    if not assert_sentence:
        raise get_my_api_error(status=status, msg=msg)
    return 1


# 获取请求头信息
def get_authorization_header(request):
    auth = request.META.get('HTTP_AUTHORIZATION', b'')
    if isinstance(auth, type('')):
        auth = auth.encode(HTTP_HEADER_ENCODING)
    return auth


# 自定义认证方式，这个是后面要添加到设置文件的
class ExpiringTokenAuthentication(BaseAuthentication):
    model = Token

    def authenticate(self, request):
        auth = get_authorization_header(request)
        if not auth:
            return None
        try:
            token = auth.decode()
        except UnicodeError:
            msg = ugettext_lazy("无效的Token， Token头不应包含无效字符")
            raise exceptions.AuthenticationFailed(msg)

        return self.authenticate_credentials(token)

    def authenticate_credentials(self, key):
        # 尝试从缓存获取用户信息（设置中配置了缓存的可以添加，不加也不影响正常功能）
        # token_cache = 'token_' + key
        # cache_user = cache.get(token_cache)
        # if cache_user:
        #     return cache_user, cache_user   # 这里需要返回一个列表或元组，原因不详
        # 缓存获取到此为止

        # 下面开始获取请求信息进行验证
        try:
            token = self.model.objects.get(key=key)
        except self.model.DoesNotExist:
            raise exceptions.AuthenticationFailed("认证失败")

        if not token.user.is_active:
            raise exceptions.AuthenticationFailed("用户被禁用")

        # Token有效期时间判断（注意时间时区问题）
        # 我在设置里面设置了时区 USE_TZ = False，如果使用utc这里需要改变。
        # if (datetime.datetime.now() - token.created) > datetime.timedelta(hours=10):
        #     raise exceptions.AuthenticationFailed('认证信息已过期')

        # 加入缓存增加查询速度，下面和上面是配套的，上面没有从缓存中读取，这里就不用保存到缓存中了
        # if token:
        #     token_cache = 'token_' + key
        #     cache.set(token_cache, token.user, 24 * 7 * 60 * 60)

        # 返回用户信息
        return token.user, token

    def authenticate_header(self, request):
        return 'Token'


class NoLoginAuthentication(BaseAuthentication):
    """
    不需要要登录也能使用
    """
    model = Token

    def process_no_token(self, auth):
        """
        处理没auth的情况, 此处不需要登录, 所以不做其它处理
        """
        pass

    def authenticate(self, request):
        auth = get_authorization_header(request)
        self.process_no_token(auth)

        try:
            token = auth.decode()
        except UnicodeError:
            msg = "无效的Token， Token头不应包含无效字符"
            raise MyValidationError(detail={
                'status': status.HTTP_417_EXPECTATION_FAILED,
                'msg': msg
            })
        return self.authenticate_credentials(token)

    def authenticate_credentials(self, key):
        try:
            tokens = self.model.objects.filter(key=key)
            if tokens.count() == 0:
                class token:
                    """
                    伪造一个token回去
                    """
                    user = AnonymousUser()  # 匿名用户
            else:
                token = tokens[0]
                if not token.user.is_active:
                    msg = "用户被禁用"
                    raise MyValidationError(detail={
                        'status': status.HTTP_403_FORBIDDEN,
                        'msg': msg
                    })
            return token.user, token
        except self.model.DoesNotExist:
            msg = "认证失败"
            raise MyValidationError(detail={
                'status': status.HTTP_401_UNAUTHORIZED,
                'msg': msg
            })

    def authenticate_header(self, request):
        return 'Token'


class MustLoginAuthentication(NoLoginAuthentication):
    """
    必须要登录才能使用
    """
    def process_no_token(self, auth):
        """
        处理没auth的情况
        """
        if not auth:
            raise MyValidationError(detail={
                'status': status.HTTP_401_UNAUTHORIZED,
                'msg': "请先登录!"
            })


from functools import wraps


def my_permission_decorator(status_code=status.HTTP_403_FORBIDDEN, msg="FORBIDDEN"):
    """
    permission装饰器

    - 如果ret为False, 则抛出一个MyValidationError
    """
    def permission_decorator(func):
        @wraps(func)
        def wrapped_function(*args, **kwargs):
            ret = func(*args, **kwargs)
            if bool(ret):
                return True
            else:
                e = get_my_validation_error(status=status_code, msg=msg)
                raise e
        return wrapped_function
    return permission_decorator






from bddjango import get_base_model
from bddjango import get_base_queryset
from pgZhSearch import search_conf_name


class BaseFullTextSearchMixin:
    """
    # 基本全文检索Mixin

    - 启用全文检索: `search_keywords`
    """
    relevant_obj_search_fields = ['A', 'B']     # 自动检索指定等级的字段
    ret_original_data_if_not_exists = True      # 返回原始数据, 如果`search`的结果为空
    _name = 'BaseFullTextSearchView'

    def get_ordered_queryset(self):
        query_dc = self.get_request_data()
        keywords = query_dc.get('search_keywords')
        if keywords:
            ret = self.queryset.search(keywords)
        else:
            ret = super().get_ordered_queryset()
        return ret


class RelevantObjectRecommendationMixin(BaseFullTextSearchMixin):
    """
    # 相关对象推荐Mixin

    - 启用全文检索: `search_keywords`
    - 指定`relevant_id`, 推荐该对象的相关对象
    """

    use_relevant_id_param = True                # 是否启用本功能
    relevant_base_model = None                  # 要推荐的模型, 例如推荐指定`人物表`相关的`事件表`, 默认为自身
    relevant_obj_search_fields = None           # 指定检索字段
    relevant_obj_search_levels = ['A', 'B']     # 若未指定`relevant_obj_search_fields`, 将自动选择指定等级的字段
    ret_original_data_if_not_exists = True      # 返回原始数据, 如果`search`的结果为空

    def get_ordered_queryset(self):
        ret = super().get_ordered_queryset()

        param_name = 'relevant_id'
        query_dc = self.get_request_data()

        if self.use_relevant_id_param and param_name in query_dc:
            relevant_id = query_dc.get(param_name)
            obj = get_base_queryset(ret).get(id=relevant_id)        # 这个得在全局检索
            ret = ret.exclude(id=relevant_id)
            qs_ls_0 = ret       # 万一没检索到相关结果, 就返回这个
            base_model = self.relevant_base_model if self.relevant_base_model else get_base_model(self.queryset)
            assert hasattr(base_model, search_conf_name), f'base_model[{base_model}]没有search_conf_name[{search_conf_name}]!'

            relevant_obj_search_fields = self.relevant_obj_search_fields
            if relevant_obj_search_fields is None:
                # 找出['A', 'B']等级的字段
                relevant_obj_search_fields = [k for k, v in getattr(base_model, search_conf_name).items() if v in self.relevant_obj_search_levels]

            search_keywords = None
            for sf in relevant_obj_search_fields:
                _value = f' {getattr(obj, sf)} '
                if search_keywords is None:
                    search_keywords = _value
                else:
                    search_keywords += _value
            ret = ret.search(search_keywords)
            if self.ret_original_data_if_not_exists and not ret.exists():
                # print(f"--- relevant_search: ret not exists! relevant_id: {relevant_id}")
                ret = qs_ls_0       # 返回相关检索之前进行过基本字段过滤的queryset
        return ret








"""
# BaseAdmin控制选项

> 这里应该用装饰器模式的... 后续改进.
> 导入功能要求用户必须拥有对应的add权限.

- stop_field_ls                                         # 停用词列表, 字段名包含该词的字段将不展示.

- move_id_to_tail = False                               # id移到最后一列去

- origin_list = False                                   # 展示原str

- `ExcelImportExportAdmin`选项
    - default_export_action = True                      # 默认增加导出按钮
    - `ExportExcelMixin`导出选项
        - export_asc = False                            # 按id升序导出
        - add_index = False                             # 导出时是否增加index列

- orm_executor = True                                   # 使用orm过滤器

- `ImportAdmin`导入导出选项
    - change_list_template = CHANGE_LIST_HTML_PATH      # 模板路径, 一般CHANGE_LIST_HTML_PATH的路径在'~bddjango/templates/entities/...'中

    - custom_import_and_export_buttons = True           # 是否显示自定义的导入导出按钮
    - has_import_perm = True                            # 导入数据
    - has_export_perm = True                            # 全部导出
    - check_import_and_export_perm = True               # 是否检查导入导出按钮的权限

"""

import csv
import datetime
import threading
import numpy as np
import pandas as pd
import xlrd
import os
from openpyxl import Workbook
from bdtime import Time
from django import forms
from django.shortcuts import render, redirect, HttpResponse
from django.urls import path
from django.contrib import admin
from django.contrib import messages
from django.utils.http import urlquote

from .. import get_model_max_id_in_db
from .. import reset_db_sequence
from ..pure import remove_temp_file

import shutil
# --- 初始化环境 ---
from .admin_env_init import CHANGE_LIST_HTML_PATH, TEMPDIR, BD_USE_GUARDIAN, CHANGE_FORM_TEMPLATE, BD_USE_SIMPLEUI
from tqdm import tqdm
from pandas._libs.tslibs.timestamps import Timestamp
from bddjango import get_base_model, get_model_max_id_in_db, old_get_model_max_id_in_db
from django.contrib.admin.helpers import ACTION_CHECKBOX_NAME
import datetime as dt
from bdtime import tt
import time
import pandas as pd
from django.contrib.auth import get_permission_codename
import re
import html


IAdmin = admin.ModelAdmin


if BD_USE_SIMPLEUI:
    from simpleui.admin import AjaxAdmin
    IAdmin = AjaxAdmin


# if BD_USE_GUARDIAN:
#     from guardian.admin import GuardedModelAdmin
#
#     IAdmin = GuardedModelAdmin
#     IAdmin.change_form_template = CHANGE_FORM_TEMPLATE


# region # --- 处理df中的特殊格式
def conv_date_field_str_format(ts):
    """
    DateField的格式转换

    - 将csv中的时间字符转为符合django的时间格式
    """
    if isinstance(ts, Timestamp):
        return ts

    if isinstance(ts, datetime.datetime):
        return ts

    if not ts or ts == 'None':
        return None

    if isinstance(ts, float) and np.isnan(ts):
        return None

    if '/' in ts:
        ts = datetime.datetime.strptime(ts, '%Y/%m/%d')
        ts = datetime.datetime.strftime(ts, '%Y-%m-%d')
    # elif '.' in ts:
    #     print('11111111')
    else:
        try:
            # 若匹配得到， 说明格式不用换
            datetime.datetime.strptime(ts, '%Y-%m-%d')
        except:
            ts = None
    return ts


def conv_date_time_field_str_format(ts):
    """
    DateTimeField的格式转换

    - 将csv中的时间字符转为符合django的时间格式
    """
    if isinstance(ts, Timestamp):
        return ts

    if not ts or ts == 'None':
        return None

    if isinstance(ts, float) and np.isnan(ts):
        return None

    if '.' in ts:
        # 毫秒级
        if '/' in ts:
            ts = datetime.datetime.strptime(ts, "%Y/%m/%d %H:%M:%S.%f")
            ts = datetime.datetime.strftime(ts, '%Y-%m-%d %H:%M:%S.%f')
        else:
            try:
                # 若匹配得到， 说明格式不用换
                datetime.datetime.strptime(ts, '%Y-%m-%d %H:%M:%S.%f')
            except:
                ts = None
        return ts

    if '/' in ts:
        ts = datetime.datetime.strptime(ts, '%Y/%m/%d %H:%M:%S')
        ts = datetime.datetime.strftime(ts, '%Y-%m-%d %H:%M:%S')
    else:
        try:
            # 若匹配得到， 说明格式不用换
            datetime.datetime.strptime(ts, '%Y-%m-%d %H:%M:%S')
        except:
            ts = None
    return ts


def format_time_column(df1, column_name):
        # 多行一起处理
        ts_ls = df1[column_name]
        ts_ls = [datetime.datetime.strptime(ts, '%Y/%m/%d') for ts in ts_ls]
        ts_ls = [datetime.datetime.strftime(ts, '%Y-%m-%d') for ts in ts_ls]
        df1[column_name] = ts_ls


def conv_nan(xx):
    # df中的特殊字符nan处理
    if xx == 'None' or (isinstance(xx, float) and np.isnan(xx)):
        return None
    else:
        return xx

# endregion


class BulkDeleteMixin:
    """
    批量删除

    - django自带的删除法太慢了, 弄个批量删除

    - django的actions文档: https://docs.djangoproject.com/en/3.2/ref/contrib/admin/actions/
    """

    @admin.action(permissions=['delete'])
    def bulk_delete(self, request, queryset=None, model=None):
        t_delete = Time()
        count = queryset.count()
        queryset.delete()
        self.message_user(request, f"成功删除{count}条数据, 耗时: {t_delete.now(1)}秒.")

    bulk_delete.short_description = "批量删除"
    bulk_delete.icon = 'el-icon-delete'
    bulk_delete.confirm = "确定要批量删除选中的数据么?"


class ExportExcelMixin:
    export_asc = False  # 按id升序导出
    add_index = False

    def export_as_excel(self, request, queryset=None, model=None, extra_fields_dc=None, add_index_column=0,
                        ordering=None):
        if model is None:
            # 如果没有指定model, 则采用默认的model, 并导出全部
            if hasattr(self, 'model'):
                base_model = self.model
            else:
                base_model = get_base_model(queryset)
            meta = base_model._meta
        else:
            meta = model._meta
            queryset = model.objects.all()

        # md = get_base_model(queryset)
        field_names = [field.name for field in meta.fields]
        verbose_names = [field.verbose_name for field in meta.fields]
        conv_name_to_verbose_dc = dict(zip(field_names, verbose_names))
        if extra_fields_dc:
            conv_name_to_verbose_dc.update(extra_fields_dc)

        if model is None:
            """
            导出全部的时候, 大文件用流媒体传输, 并分批次读入, 防止超过云服务器带宽.
            未完待续...
            """
            from django.http import StreamingHttpResponse
            from django_pandas import io
            from bdtime import tt
            from django.db import models as m
            # print('~~~~~~~~~~~~ queryset.count():', queryset.count())
            qs: m.QuerySet = queryset

            tt.__init__()

            # print('开始读入sql...')
            qs: m.QuerySet
            # import pandas as pd
            # from django.db import connection
            # df = pd.concat([df for df in pd.read_sql(f'SELECT * FROM {meta.db_table}', connection, chunksize=100)], axis=0)
            df = io.read_frame(qs=qs)
            df.columns = [conv_name_to_verbose_dc.get(col) if conv_name_to_verbose_dc.get(col) else col for col in
                          df.columns]
            # print('df读入成功: ', tt.now(2))

            if add_index_column:
                # 增加一个序号列, start为 add_index_column - 1
                index = pd.Series(df.index) + int(add_index_column) - 1  # add_index_column为2, 则从1开始
                df.insert(loc=0, column='序号', value=index, allow_duplicates=False)

            if ordering:
                df = df[ordering]

            # print('开始保存为excel...', tt.now(2))
            time_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            tmp_file_name = f"{meta.verbose_name}_{time_str}.xlsx"
            fpath = os.path.join(TEMPDIR, tmp_file_name)
            # df.to_excel(fpath, index=False, encoding='unicode_escape')       # unicode_escape
            df.to_excel(fpath, index=self.add_index)  # unicode_escape
            # print('保存为Excel成功!', tt.now(2))
            del df

            # big file download
            def file_iterator(file_name, open_model='rb', chunk_size=512):
                with open(file_name, open_model) as f:  # , encoding=encoding
                    while True:
                        c = f.read(chunk_size)
                        if c:
                            yield c
                        else:
                            break

            # print('write response', tt.now(2))
            response = StreamingHttpResponse(file_iterator(fpath))
            # response['Content-Type'] = 'application/octet-stream'
            filename = urlquote(f"{meta.verbose_name}.xlsx")
            response['Content-Type'] = 'application/msexcel'
            response['Content-Disposition'] = 'attachment;filename="{0}"'.format(filename)

            print('return response', tt.now(2))
            return response
        else:
            response = HttpResponse(content_type='application/msexcel')
            filename = urlquote(f"{meta.verbose_name}.xlsx")
            response['Content-Disposition'] = f'attachment; filename={filename}'
            wb = Workbook()
            ws = wb.active

            if self.export_asc:
                if queryset.count() and hasattr(queryset[0], 'id'):
                    queryset = queryset.order_by('id')
            ws.append(verbose_names)

            total = queryset.count()
            tq = tqdm(total=total)
            for obj in queryset:
                tq.update(1)
                data = [f'{getattr(obj, field)}' for field in field_names]
                try:
                    ws.append(data)
                except Exception as e:
                    print(data)
                    raise e
            wb.save(response)
            return response

    export_as_excel.short_description = "导出所选数据"
    # export_as_excel.acts_on_all = True
    # export_as_excel.type = 'success'
    # export_as_excel.icon = 'el-icon-upload'
    export_as_excel.icon = 'el-icon-download'


class ExportCsvMixin:
    def export_as_csv(self, request, queryset, model=None):
        if model is None:
            meta = self.model._meta
        else:
            meta = model._meta
        field_names = [field.name for field in meta.fields]
        verbose_names = [field.verbose_name for field in meta.fields]

        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename={}.csv'.format(meta)
        writer = csv.writer(response)
        writer.writerow(verbose_names)
        for obj in queryset:
            writer.writerow([getattr(obj, field) for field in field_names])
        return response

    export_as_csv.short_description = "  导出选中数据"  # 有图标的话要空俩格, 不然太近了
    # export_as_csv.acts_on_all = True
    # export_as_csv.icon = 'fas fa-audio-description'
    # export_as_csv.icon = 'fas fa-download'
    export_as_csv.icon = 'fas fa-download'


class CsvImportForm(forms.Form):
    csv_file = forms.FileField()


def _get_base_admin_dc_by_iadmin(iadmin):       # 类似装饰器模式, 通过IAdmin来获取装饰后的BaseAmin
    if BD_USE_GUARDIAN:
        from guardian.admin import GuardedModelAdminMixin

        class iadmin(GuardedModelAdminMixin, iadmin):
            change_form_template = CHANGE_FORM_TEMPLATE

    class IDAdmin(iadmin):
        """
        * 保存时自动处理id, 解决postgre在批量导入数据后的主键冲突问题.

        - 该方法仅对admin界面手动保存时调用的save_model()方法生效, 不影响obj.save()方法效率.
        """

        def save_model(self, request, obj, form, change):
            """
            参数change分辨保存or修改.
            若为保存, 则model的id值自动更新为数据库中最大id+1.
            """
            # if change is False:
            #     # meta = obj._meta
            #     # obj.id = get_model_max_id_in_db(model=None, meta=meta)
            #     obj.id = get_model_max_id_in_db(model=obj)
            # obj.save()

            try:
                obj.save()
            except Exception as e:
                msg = f'可能为pgsql的id引起的错误:' + str(e)

                print(msg)
                reset_db_sequence(obj)
                # obj.id = get_model_max_id_in_db(get_base_model(obj))
                obj.save()

    class ImportAdmin(IDAdmin):
        """
        导入类, CSV和Excel通用

        - 不能与admin.ModelAdmin一起用!
        """
        change_list_template = CHANGE_LIST_HTML_PATH

        custom_import_and_export_buttons = True
        has_import_perm = True      # 导入数据
        has_export_perm = True      # 全部导出
        check_import_and_export_perm = True     # 是否检查导入导出按钮的权限

        def import_csv(self, request):
            t_import = Time()
            index = 0
            try:
                if request.method == "POST":
                    csv_file = request.FILES.get("csv_file")
                    assert csv_file, '文件不能为空!'
                    assert csv_file._name and csv_file._name.__contains__('.'), '文件名不能为空, 且必须有后缀名!'
                    f_format = csv_file._name.rsplit('.', 1)[-1]
                    format_ls = ['xls', 'xlsx', 'csv']
                    assert f_format in format_ls, f'不支持的文件类型! 目前仅支持{format_ls}.'

                    read_data = csv_file.read()

                    time_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

                    tempdir = 'tempdir'
                    if not os.path.exists(tempdir):
                        os.mkdir(tempdir)

                    if f_format == 'csv':
                        try:
                            encoding = 'utf-8'
                            file_data = read_data.decode(encoding)
                        except Exception as e:
                            print('-- 尝试用gbk编码 --')
                            encoding = 'gbk'
                            file_data = read_data.decode(encoding)

                        fname = f'f_{time_str}.csv'
                        fname = os.path.join(tempdir, fname)

                        with open(fname, 'w', encoding=encoding) as f:
                            f.write(file_data)

                        # 为解决字段内有逗号导致分割错误问题, 只能采用pd了
                        df = pd.read_csv(fname, encoding=encoding)
                    elif f_format in ['xlsx', 'xls']:
                        wb = xlrd.open_workbook(file_contents=read_data)
                        df = pd.read_excel(wb)
                    else:
                        df = None

                    df_rows = df.shape[0]        # 一共多少行数据

                    # 删除Unnamed列
                    df = df.loc[:, ~df.columns.str.match('Unnamed')]

                    # 找出在model定义的列
                    meta = self.model._meta
                    field_names = [field.name for field in meta.fields]
                    verbose_names = [field.verbose_name for field in meta.fields]
                    field_dc = dict(zip(verbose_names, field_names))
                    valid_columns = [column_i in field_names or column_i in verbose_names for column_i in df.columns]
                    df = df.loc[:, valid_columns]
                    titles = df.columns.tolist()
                    title_ls = [field_dc.get(title_i) if field_dc.get(title_i) else title_i for title_i in titles]
                    curr_id = get_model_max_id_in_db(self.model)

                    df1 = df.copy()
                    df1.columns = title_ls

                    # 这里有问题, 服务器带宽太小的话, 一次性bulk_update会造成tcp拥塞, 从而导致服务器瘫痪!
                    # 必须仿照navicat的处理方式, 设置batch_size, 小批量处理数据.
                    # 同时出现问题, 如何让用户看到进度条? 用异步的方式, 或者websocket?
                    md_ls = []
                    my_tqdm = tqdm(total=df_rows)
                    for index, row in df1.iterrows():
                        content_ls = row.values.tolist()

                        # 处理DateField字段
                        for i in range(len(title_ls)):
                            title_i = title_ls[i]
                            content_ls[i] = conv_nan(content_ls[i])

                            attr = getattr(self.model, title_i)
                            if not hasattr(attr, 'field'):
                                continue
                            attr_field_name = attr.field.__class__.__name__

                            if attr_field_name in ['DateField', 'DateTimeField']:
                                # print(i, title_i, content_ls[i])
                                if attr_field_name == 'DateTimeField':
                                    res = conv_date_time_field_str_format(ts=content_ls[i])
                                    # if res is None:
                                    #     print(content_ls[i])
                                    content_ls[i] = res
                                else:
                                    ts = content_ls[i]
                                    res = conv_date_field_str_format(ts=ts)
                                    # if res is None:
                                    #     print(content_ls[i])
                                    content_ls[i] = res

                        dc = dict(zip(title_ls, content_ls))
                        dc.update({'id': curr_id})
                        curr_id += 1
                        md = self.model(**dc)
                        md_ls.append(md)
                        my_tqdm.update(1)
                        # self.model.objects.create(**dc)

                    self.model.objects.bulk_create(md_ls,batch_size=100)
                    reset_db_sequence(self.model)
                    self.message_user(request, f"{f_format}文件导入成功! 一共导入{df_rows}条数据, 耗时: {t_import.now(1)}秒.")
                    self.remove_temp_file(tempdir)

                    return redirect("..")
            except Exception as e:
                self.message_user(request, f"第 {index+1} 条数据导入失败!</br>错误信息：&nbsp;" + str(e), level=messages.ERROR)
                return redirect("..")

            form = CsvImportForm()
            payload = {"form": form}
            return render(request, "admin/csv_form.html", payload)

        def bulk_save(self, request, md_ls, f_format, df_rows, t_import, tempdir, batch_size=100):
            def _bulk_save():
                self.model.objects.bulk_create(md_ls, batch_size=batch_size)
                reset_db_sequence(self.model)
                self.message_user(request, f"{f_format}文件导入成功! 一共导入{df_rows}条数据, 耗时: {t_import.now(1)}秒.")
                self.remove_temp_file(tempdir)

            if df_rows < 3000:
                return _bulk_save()
            else:
                print('~~~ 多线程_bulk_save')
                self.message_user(request, f"{f_format}文件的{df_rows}条数据格式正确, 正在后台导入中, 请稍后查看...")
                t1 = threading.Thread(target=_bulk_save, args=())
                t1.start()

        def export_all_csv(self, request):
            # self.message_user(request, "成功导出全部数据为csv文件")
            return ExportCsvMixin().export_as_csv(request, queryset=self.model.objects.all(), model=self.model)

        def export_all_excel(self, request):
            # self.message_user(request, "成功导出全部数据为csv文件")
            ret = ExportExcelMixin().export_as_excel(request, queryset=self.model.objects.all(), model=self.model)
            self.remove_temp_file(TEMPDIR)
            return ret

        def get_urls(self):
            my_urls = [
                path('import-csv/', self.import_csv),
                path('export_all_csv/', self.export_all_csv),
                path('export_all_excel/', self.export_all_excel),
            ]
            return my_urls + super().get_urls()

        def remove_temp_file(self, tempdir):
            return remove_temp_file(tempdir)

        def changelist_view(self, request, extra_context=None):
            if not self.check_import_and_export_perm:
                has_import_perm = has_export_perm = True
            else:
                opts = self.opts

                def has_action_permission(opts, action):
                    codename = get_permission_codename(action, opts)
                    has_perm = request.user.has_perm('%s.%s' % (opts.app_label, codename))
                    return has_perm

                has_add_perm = has_action_permission(opts, 'add')
                has_change_perm = has_action_permission(opts, 'change')
                has_view_perm = has_action_permission(opts, 'view')

                has_import_perm = self.has_import_perm and has_add_perm and has_change_perm
                has_export_perm = self.has_export_perm and has_view_perm

            if extra_context is None:
                extra_context = {}
            extra_context.update({
                'custom_import_and_export_buttons': self.custom_import_and_export_buttons,
                'has_import_perm': has_import_perm,
                'has_export_perm': has_export_perm,
            })

            ret = super().changelist_view(request, extra_context=extra_context)
            return ret

    class CsvImportExportAdmin(ImportAdmin, ExportCsvMixin):
        """
        CSV导入/导出Admin类

        - 不能与admin.ModelAdmin一起用!
        """

        actions = ['export_as_csv']

    class ExcelImportExportAdmin(ImportAdmin, ExportExcelMixin):
        """
        Excel导入/导出Admin类

        - 不能与admin.ModelAdmin一起用!
        """
        default_export_action = True        # 默认增加导出按钮

        def get_actions(self, request):
            ACTION_NAME = 'export_as_excel'
            if self.default_export_action and ACTION_NAME not in self.actions:
                if self.actions is None:
                    self.actions = []
                self.actions.append(ACTION_NAME)
            ret = super().get_actions(request)
            return ret

    class ListDisplayAdmin(ExcelImportExportAdmin):
        """
        * admin展示界面

        - 展示所有字段, 默认前两列为点击链接. 并去除'stop_field_ls'中的字段

        - stop_field_ls: 停用词列表, 字段名包含该词的字段将不展示.
            - 注意url和href的区别: url为本地路由; href超链接, 不显示
        """
        list_display = '__all__'        # 默认展示全部
        stop_field_ls = []              # 停用字段
        move_id_to_tail = False      # id移到最后一列去

        origin_list = False     # 展示原str

        def __init__(self, *args, **kwargs):
            """
            增加一个可点击字段
            """
            super().__init__(*args, **kwargs)

            meta = self.model._meta
            field_names = [field.name for field in meta.fields]

            if self.list_display == [] or self.list_display is None:
                self.list_display = ('__str__', )
                self.origin_list = True
                return

            if not self.list_display_links and len(field_names) >= 2:
                if field_names[0] == 'id':
                    self.list_display_links = ('id', field_names[1])

            if (isinstance(self.list_display, str) and self.list_display == '__all__') or '__all__' in self.list_display:
                res = []
                for f in field_names:
                    if f not in self.stop_field_ls:
                        res.append(f)
                self.list_display = res

            return

        def get_list_display(self, request):
            ret = super().get_list_display(request)
            meta = self.model._meta
            field_names = [field.name for field in meta.fields]

            # model_str = self.model.__str__(self.model)
            # if not self.origin_list and ret == ('__str__',) and isinstance(model_str, str) and model_str.__contains__('ModelBase'):
            #     """如果用户没对list_display进行变更的话, 则自定义list_display字段"""
            #     meta = self.model._meta
            #     field_names = [field.name for field in meta.fields]
            #
            #     ret = []
            #     for f in field_names:
            #         if isinstance(f, str):
            #             # 如果有停用词, 则不加入ret中
            #             flag = False
            #             for r in self.stop_field_ls:
            #                 if f.__contains__(r):
            #                     flag = True
            #                     break
            #             if flag:
            #                 continue
            #         ret.append(f)

            if self.move_id_to_tail and 'id' in field_names and id not in self.stop_field_ls:
                ret.remove('id')
                ret.append('id')
            return ret

    class ForceRunActionsAdmin(ListDisplayAdmin):
        """
        增加强制运行actions功能
        """
        def changelist_view(self, request, extra_context=None):
            """
            这里想跳过"必须选择一个数据"的确认框,
            但django原生admin可以, 而simpleui无法实现,
            经过改进后, 以"fc_"开头的方法将强制运行

            - https://stackoverflow.com/questions/4500924/django-admin-action-without-selecting-objects
            """
            MyModel = self.model

            # 强制运行以"fc_"开头的action
            if 'action' in request.POST and request.POST['action'].startswith('fc_'):
                if not request.POST.getlist(ACTION_CHECKBOX_NAME):
                    post = request.POST.copy()
                    for u in MyModel.objects.all():
                        post.update({ACTION_CHECKBOX_NAME: str(u.id)})
                    request._set_post(post)
            return super().changelist_view(request, extra_context)

    class BaseAdmin(ForceRunActionsAdmin):
        """
        # 若search_term以变量prefix的值开头, 则检索最近xx条记录.
            - 如'~10'代表检索最近10条记录

        # 支持orm检索
        > 使用反引号[`]来替代引号["]
            - .filter(id__lt=20, wiki_id="asd")
            - .filter(wiki_id=`asd`).order_by(`-id`)[:3]
        """
        _tmp = None
        orm_executor = True  # 使用orm过滤器

        _ajax_search_term = None  # 用来ajax过滤
        _ajax_return_qs_ls = None  # 以便使用ajax方法返回的qs_ls

        def get_search_results(self, request, queryset, search_term):
            if not self.orm_executor:
                ret = super().get_search_results(request, queryset, search_term)
                return ret

            # --- 根据ajax方法返回qs_ls
            bd_ajax_return_qs_ls = '_ajax_return_qs_ls'
            if hasattr(self, bd_ajax_return_qs_ls) and getattr(self, bd_ajax_return_qs_ls) is not None:
                queryset = getattr(self, bd_ajax_return_qs_ls)
                setattr(self, bd_ajax_return_qs_ls, None)
                ret = (queryset, False)
                return ret

            # --- 根据orm语句进行过滤
            prefix = '~'  # 反向过滤~pk

            reg = re.compile(r'^\..*?[\)\]]$')  # 让.filter(xxx), .order_by(xxx)等orm语句可以执行

            if hasattr(self, '_ajax_search_term') and getattr(self, '_ajax_search_term'):
                search_term = self._ajax_search_term
                self._ajax_search_term = None

            if search_term and reg.match(search_term):
                print('search_term: ', search_term)

                # html转义
                search_term = search_term.replace('`', '"')
                while ('=&' in search_term):     search_term = html.unescape(search_term)
                try:
                    self._tmp = queryset
                    exec(f'self._tmp = self._tmp{search_term}')
                    queryset = self._tmp
                    # exec(f'queryset = queryset{search_term}')     # 局部变量无法用exec赋值!
                except Exception as e:
                    msg = f'orm语句执行出错! <br> search_term: {search_term},<br>  error: {e}'
                    self.message_user(request, msg)
                ret = (queryset, False)
            elif search_term.startswith(prefix):
                try:
                    if search_term.startswith(prefix):
                        search_term = search_term
                        search_term = int(search_term[len(prefix):])
                except Exception as e:
                    raise TypeError(f'pk必须为整数! {e}')
                id_qsv_ls = queryset.values('pk')[:search_term]
                queryset = queryset.filter(pk__in=id_qsv_ls)
                ret = (queryset, False)
            else:
                ret = super().get_search_results(request, queryset, search_term)

            return ret

    ret = {
        'IDAdmin': IDAdmin,
        'ImportAdmin': ImportAdmin,
        'CsvImportExportAdmin': CsvImportExportAdmin,
        'ExcelImportExportAdmin': ExcelImportExportAdmin,
        'ListDisplayAdmin': ListDisplayAdmin,
        'ForceRunActionsAdmin': ForceRunActionsAdmin,
        'BaseAdmin': BaseAdmin,
    }
    return ret


from . import adminclass


# decorator, 装饰器模式, 给admin_class附加额外的功能
def get_base_admin_dc_by_iadmin(iadmin=IAdmin, key='BaseAdmin') -> adminclass.BaseAdmin:
    admin_dc = _get_base_admin_dc_by_iadmin(iadmin)
    admin_class = admin_dc.get(key)
    return admin_class


if BD_USE_SIMPLEUI:
    MyAjaxAdmin: adminclass.BaseAdmin = get_base_admin_dc_by_iadmin(AjaxAdmin)


class CustomIAdmin:
    def __init__(self, iadmin):
        self.iadmin = iadmin

    def get_base_admin_dc_by_iadmin(self, key):
        return get_base_admin_dc_by_iadmin(key, iadmin=self.iadmin)

    @property
    def IDAdmin(self)->adminclass.IDAdmin:
        ret = self.get_base_admin_dc_by_iadmin('IDAdmin')
        return ret

    @property
    def BaseAdmin(self) -> adminclass.BaseAdmin:
        ret = self.get_base_admin_dc_by_iadmin('BaseAdmin')
        return ret

    @property
    def CsvImportExportAdmin(self):
        ret = get_base_admin_dc_by_iadmin('CsvImportExportAdmin')
        return ret





from django.conf import settings


# --- 默认分页器参数设置 PAGINATION_SETTINGS
PAGINATION_SETTINGS = {
    'page_size': 10,       # 每页16个
    'page_size_query_param': 'page_size',       # 前端控制每页数量时使用的参数名, 'page_size'
    'page_query_param': 'p',        # 页码控制参数名"p"
    'max_page_size': 1000,      # 最大1000页
}

if hasattr(settings, 'PAGINATION_SETTINGS'):
    PAGINATION_SETTINGS.update(settings.PAGINATION_SETTINGS)







"""
纯python的功能函数
"""
import json
import pandas as pd
import inspect
import functools
import os
import threading
import shutil
from bdtime import Time
import datetime as dt


TEMPDIR = 'tempdir'     # 临时文件夹


def add_status_and_msg(dc_ls, status=200, msg=None):
    if status != 200 and msg is None:
        msg = '请求数据失败!'

    if status == 200 and msg is None:
        msg = "ok"

    ret = {
        'status': status,
        'msg': msg,
        'result': dc_ls
    }
    return ret


def show_json(data: dict, sort_keys=False):
    try:
        print(json.dumps(data, sort_keys=sort_keys, indent=4, separators=(', ', ': '), ensure_ascii=False))
    except:
        if isinstance(data, dict):
            for k, v in data.items():
                print(k, ' --- ', v)
        else:
            for k, v in data:
                print(k, ' --- ', v)


def show_ls(data: list, ks=None):
    for dc in data:
        if ks:
            if isinstance(ks, str):
                ks = [ks]
            d = [dc.get(k) for k in ks]
        else:
            d = dc
        print(d)


def add_space_prefix(text, n, more=True, prefix='\u3000'):
    text = str(text)
    if more:
        ret = prefix * n + text
    else:
        ret = prefix * (n - len(text)) + text
    return ret


def create_file_if_not_exist(file_name):
    if not os.path.exists(file_name):
        with open(file_name, 'w') as f:
            f.write('')
        return False
    return True


def create_dir_if_not_exist(dirpath: str):
    if not os.path.exists(dirpath):
        os.mkdir(dirpath)
        return False
    return True


def get_class_that_defined_method(meth):
    """
    get mehod's class
    """
    if isinstance(meth, functools.partial):
        return get_class_that_defined_method(meth.func)
    if inspect.ismethod(meth) or (inspect.isbuiltin(meth) and getattr(meth, '__self__', None) is not None and getattr(meth.__self__, '__class__', None)):
        for cls in inspect.getmro(meth.__self__.__class__):
            if meth.__name__ in cls.__dict__:
                return cls
        meth = getattr(meth, '__func__', meth)  # fallback to __qualname__ parsing
    if inspect.isfunction(meth):
        cls = getattr(inspect.getmodule(meth),
                      meth.__qualname__.split('.<locals>', 1)[0].rsplit('.', 1)[0],
                      None)
        if isinstance(cls, type):
            return cls
    return getattr(meth, '__objclass__', None)  # handle special descriptor objects


def get_whole_codename_by_obj_and_perm(obj=None, perm=None, suffix_model_name=False):
    """
    得到obj的perm对应的完整codename: whole_codename

    - eg:
    ```
    perm = get_whole_codename_by_obj_and_perm(obj=model, perm=perm_codename, suffix_model_name=ssuffix_model_name)
    ret = user.has_perm(perm)
    ```

    :param obj: 模型 or 对象
    :param perm: 权限名
    :param suffix_model_name: perm里边没有obj对应model的model_name, 需要函数手动添加
    :return:
    """
    if obj:
        if suffix_model_name:
            ret = f'{obj._meta.app_label}.{perm}_{obj._meta.model_name}'
        else:
            ret = f'{obj._meta.app_label}.{perm}'
    else:
        ret = perm
    return ret


def conv_df_to_serializer_data(df) -> list:
    assert isinstance(df, pd.DataFrame), 'df的类型必须是DataFrame!'
    ret_ls = []
    for index, row in df.iterrows():
        k = row.index.tolist()
        v = row.values.tolist()
        data = dict(zip(k, v))

        ret_ls.append(data)
    return ret_ls


def convert_query_parameter_to_bool(query_parameter, false_ls=None):
    """
    将请求参数转化为`bool`类型

    :param query_parameter: 请求参数
    :param false_ls: 将转换为`false`的值
    :return: bool, true or false
    """
    if false_ls is None:
        false_ls = ['0', 0, None, 'None', 'Null', [], [''], {}, 'False', 'false', '', 'null']
    ret = query_parameter not in false_ls
    return ret


def _remove_temp_file(tempdir=TEMPDIR, MAX_TEMPS=5, desc='---', remain_rows=None, option_model='getatime', quiet=False):
    """
    清理缓存, 清空tempdir下的所有文件
    :param tempdir: 文件路径
    :param MAX_TEMPS: 最多缓存文件数量
    :param remain_rows: 最清理时留下的缓存文件数量, 如空, 则保留1/3
    :param option_model: 操作模式, os.path的[getatime, getctime, getmtime]函数

    # :param MAX_SPACE: 最大缓存文件空间
    """
    fpath_ls = os.listdir(tempdir)
    temps = len(fpath_ls)

    if temps < MAX_TEMPS:
        if not quiet:
            print(f'...缓存还足够, 不用清理... 缓存容量: {temps}/{MAX_TEMPS}')
        return False

    # --- 按option_model选择的时间函数来清理缓存文件
    if option_model in ['getatime', 'getctime', 'getmtime']:        # remain_recent
        # print(option_model)
        col_0 = 'filename'
        fpath_df = pd.DataFrame(fpath_ls, columns=[col_0])
        f = getattr(os.path, option_model)
        fpath_df['abs_fpath'] = [os.path.join(tempdir, f_i) for f_i in fpath_df[col_0]]
        fpath_df['t_tamp_ls'] = [f(os.path.join(tempdir, f_i)) for f_i in fpath_df[col_0]]
        fpath_df['t_str_ls'] = [dt.datetime.fromtimestamp(f_i).strftime("%Y-%m-%d %H:%M:%S") for f_i in fpath_df['t_tamp_ls']]

        # 删到remain_rows个文件为止
        remain_rows = remain_rows if remain_rows else MAX_TEMPS//3
        delete_rows = temps - remain_rows

        # delete_fpath_df[['t_tamp_ls', 't_str_ls']]
        delete_fpath_df = fpath_df.sort_values(by='t_tamp_ls')[:delete_rows]

        # --- 按空间清理
        # remain_fpath_df = fpath_df.sort_values(by='t_tamp_ls', ascending=False)[:MAX_TEMPS]
        # remain_fpath_df['size'] = [os.path.getsize(os.path.join(tempdir, f_i)) for f_i in remain_fpath_df[col_0]]
        #
        # remain_fpath_df[['t_str_ls', 'size']]
        # # MAX_SPACE = 1024 * 10
        #
        # s = 0
        # size_ls = fpath_df['size'].tolist()
        # size_ls.reverse()
        # for i in range(fpath_df.shape[0]):
        #     size = size_ls[i]
        #     s += size
        #     print(i, size, s)
        #     if s >= MAX_SPACE:
        #         break
        # space_i = i - 1
        # space_i
        # fpath_df.sort_values(by='size')[:MAX_TEMPS]

        fpath_ls = delete_fpath_df[col_0].tolist()
        temps = len(fpath_ls)

    tt = Time()

    tt.sleep(1)
    if not quiet:
        print(f'*************** 开始清理缓存 {tempdir} *************')
    for fpath in fpath_ls:
        i = 0
        tt.__init__()
        while tt.during(5):
            i += 1
            dirpath = os.path.join(tempdir, fpath)

            try:
                if os.path.isdir(dirpath):
                    # os.removedirs(dirpath)
                    shutil.rmtree(dirpath)
                else:
                    os.remove(dirpath)
                if not quiet:
                    print(f"~~~ success: 移除文件[{dirpath}]成功! -- 第[{i}]次")
                break
            except:
                print(f"** 第[{i}]次移除文件[{dirpath}]失败...可能文件被占用中?")
                tt.sleep(1)
                if i > 3:
                    print(f"======== Warning: 移除文件[{dirpath}]失败!")
    if not quiet:
        print(f'*************** [{desc}] 缓存清理完毕 *************')
    return True


def remove_temp_file(tempdir=TEMPDIR, MAX_TEMPS=5, desc='---', remain_rows=None, option_model='getatime', quiet=True):
    """
    当大于MAX_TEMPS时启动临时文件清理程序
    """
    temps = len(os.listdir(tempdir))

    if temps > MAX_TEMPS:
        t1 = threading.Thread(target=_remove_temp_file, args=(tempdir, MAX_TEMPS, desc, remain_rows, option_model, quiet))
        t1.start()
        return True
    else:
        if not quiet:
            print(f'{desc} --- 缓存还足够, 不用清理... 缓存文件: {temps}/{MAX_TEMPS}, tempdir: {tempdir}')
        return False


class SetUtils:
    @staticmethod
    def get_ls_a_and_b(ls_a, ls_b):
        # 获得列表a和b的交集 a and b
        ret = [j for j in ls_a if j in ls_b]
        return ret

    @staticmethod
    def get_ls_a_or_b(ls_a, ls_b):
        # 获得列表a和b的交集 a and b
        ret = list(set(ls_a) | (set(ls_b)))
        return ret

    @staticmethod
    def get_ls_a_sub_b(ls_a, ls_b):
        # 获得列表a和b的交集 a and b
        ret = list(set(ls_a) - set(ls_b))
        return ret


set_utils = SetUtils()


# def create_db_if_not_exist(pg_conf, use_zhparser=True, debug=False):
#     from psycopg2 import connect
#     from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT
#
#     assert 'postgresql' in pg_conf.get('ENGINE'), '本方案仅支持postgresql数据库!'
#
#     db_name = pg_conf.get('NAME')       # 想新建的数据库名
#
#     user = pg_conf.get('USER')
#     pwd = pg_conf.get('PASSWORD')
#     port = pg_conf.get('PORT')
#     host = pg_conf.get('HOST')
#
#     conn = connect(database="postgres", port=port, host=host, user=user, password=pwd)
#
#     conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
#
#     cursor = conn.cursor()
#
#     # 创建数据库
#     sql = f"SELECT u.datname FROM pg_catalog.pg_database u where u.datname='{db_name}';"
#     cursor.execute(sql)
#
#     row = cursor.fetchall()
#     if row:
#         msg = f'已存在数据库[{db_name}]!'
#         if debug:
#             print(msg)
#     else:
#         sql = "CREATE DATABASE {};".format(db_name)     # pg不支持 CREATE DATABASE XXX IF NOT EXISTS;
#         cursor.execute(sql)
#
#         msg = f'数据库[{db_name}]创建成功.'
#         print(msg)
#
#     cursor.close()
#     return msg
#
#
# def is_exist_table_in_db(pg_conf, table_name, debug=False):
#     """
#     判断db中是否已经迁移好了表table_name
#     """
#     from psycopg2 import connect
#
#     assert 'postgresql' in pg_conf.get('ENGINE'), '本方案仅支持postgresql数据库!'
#
#     db_name = pg_conf.get('NAME')       # 想新建的数据库名
#     user = pg_conf.get('USER')
#     pwd = pg_conf.get('PASSWORD')
#     port = pg_conf.get('PORT')
#     host = pg_conf.get('HOST')
#
#     conn = connect(database=db_name, port=port, host=host, user=user, password=pwd)
#     cursor = conn.cursor()
#
#     # 判断表是否存在
#     sql = f"SELECT count(*) FROM {table_name};"
#
#     try:
#         exist_flag = True
#         cursor.execute(sql)
#     except Exception as e:
#         exist_flag = False
#         if debug:
#             print('*** is_exist_table_in_db error: ', e)
#
#     if debug:
#         print('--- is_exist_table_in_db:', exist_flag)
#     cursor.close()
#     return exist_flag
#
#
# def judge_db_is_migrating():
#     import sys
#     if 'makemigrations' in sys.argv or 'migrate' in sys.argv:
#         return True
#     else:
#         return False



"""
分割大文件pdf
"""

import PyPDF2
import os


def get_file_size(file_path, dround=3):
    return round(os.path.getsize(file_path) / 1024 / 1024, dround)


def get_pdf_info(file_path):
    pdfFile = open(file_path, 'rb')
    pdfReader = PyPDF2.PdfFileReader(pdfFile, strict=False)
    total_pages = pdfReader.numPages
    size = get_file_size(file_path)
    return total_pages, size


def extract_pdf(origFileName, newFileName, page=None, nums=1, rotation=0):
    pdfFile = open(origFileName, 'rb')
    pdfReader = PyPDF2.PdfFileReader(pdfFile, strict=False)
    pdfWriter = PyPDF2.PdfFileWriter()

    assert page is not None, '请指定页数page!'
    page = int(page)
    for p in range(page, page + nums):
        pageObj = pdfReader.getPage(p)
        pageObj.rotateClockwise(rotation)
        pdfWriter.addPage(pageObj)

    newFile = open(newFileName, 'wb')
    pdfWriter.write(newFile)

    pdfFile.close()
    newFile.close()


def main():
    origFileName = 'p1.pdf'
    rotation = 0
    page = 1
    nums = 1
    newFileName = f'page_{page}_to_{page + nums - 1}.pdf'

    extract_pdf(origFileName, newFileName, page=page, nums=nums, rotation=rotation)


if __name__ == "__main__":
    import os
    os.chdir('myutils')
    main()


"""
依赖django的功能函数
"""

from .. import pure

import math
import pandas as pd

from django.core.paginator import Paginator
from rest_framework.response import Response
from rest_framework.pagination import PageNumberPagination
from rest_framework.generics import GenericAPIView
from rest_framework.mixins import ListModelMixin, RetrieveModelMixin
from rest_framework.request import Request
from functools import wraps
from django.db import connection
from django.db.models import Q, F
from django.db.models import QuerySet
from django.db.models import Max
from django.conf import settings
from warnings import warn
from django.core.management.color import no_style
from django.db import connection

from rest_framework.renderers import JSONRenderer
from django.forms import model_to_dict
from django.db.models import QuerySet
from rest_framework.mixins import CreateModelMixin, UpdateModelMixin, DestroyModelMixin
from rest_framework import status
from django.db.models import Model
from rest_framework.exceptions import APIException
from rest_framework.exceptions import ErrorDetail
import re
from django.contrib.contenttypes.models import ContentType
import json
from rest_framework import serializers as s
from django.http.request import QueryDict
from .auth import get_my_api_error, my_api_assert_function
import re
import sys
from .conf import PAGINATION_SETTINGS
from django.db import models as m
from django.db import connection
from copy import deepcopy
# from rest_framework import serializers


def get_list(query_dc, key):
    if isinstance(query_dc, QueryDict):
        ret = query_dc.getlist(key)
    elif isinstance(query_dc, dict):
        ret = query_dc.get(key)
    else:
        raise TypeError('query_dc类型不明!')

    if isinstance(ret, str):
        ret = [ret]

    if ret == ['']:
        ret = []

    return ret


def get_key_from_request_data_or_self_obj(request_data, self_obj, key, get_type=None):
    """
    优先检索request_data是否有key, 其次检索self_obj是否有key这个属性
    :param key: 变量名
    :return:
    """
    query_dc = request_data

    if get_type == 'list':
        value = get_list(query_dc, key)
    elif get_type == 'bool':
        value = query_dc.get(key)
        ret_0 = getattr(self_obj, key) if hasattr(self_obj, key) else None
        ret_1 = pure.convert_query_parameter_to_bool(value) if value else None
        ret = ret_1 if ret_1 is not None else ret_0
        return ret
    else:
        value = query_dc.get(key)

    # 让request携带的数据可以覆盖自身的key值
    ret_0 = getattr(self_obj, key) if hasattr(self_obj, key) else None
    ret = value if value is not None else ret_0
    return ret


def set_query_dc_value(query_dc: (QueryDict, dict), new_dc: dict):
    assert isinstance(query_dc, QueryDict), 'query_dc必须是QueryDict类型!'
    query_dc._mutable = True
    for key, value in new_dc.items():
        ls = value if isinstance(value, (tuple, list)) else [value]
        query_dc.setlist(key, ls)
    query_dc._mutable = False
    return query_dc


def get_ls_a_different_ls_b(a, b, keep_sort=True):
    """
    列表a和列表b的差集a-b, 并保留a列表原顺序
    """
    ret = list(set(a) - set(b))
    if keep_sort:
        ret.sort(key=a.index)
    return ret


def get_field_names_by_model(model_class, field_attr='name', exclude_fields=None):
    """
    获得`model_class.meta`中的字段名属性

    :param model_class: 目标模型
    :param field_attr: 可取[name, verbose_name]
    :param exclude_fields: 要排除的字段       # 差集 --> set(a).difference(set(b)), DRF在Meta中用exclude指定
    :return:
    """
    if isinstance(model_class, m.QuerySet):
        model_class = get_base_model(model_class)
    assert hasattr(model_class, '_meta'), f'model_class没有`_meta`属性? type(model_class): {type(model_class)}'

    fields = model_class._meta.fields
    field_names = [getattr(field, field_attr) for field in fields]

    if exclude_fields:
        assert isinstance(exclude_fields, (list, tuple)), '`exclude_fields`必须为list或tuple类型!'
        field_names = get_ls_a_different_ls_b(field_names, exclude_fields, keep_sort=True)
    return field_names


def get_base_serializer(base_model, base_fields='__all__', auto_generate_annotate_fields=None):
    """
    生成一个基础序列化器

    :param base_model: queryset或者base_model
    :param base_fields: 想要返回的字段
    :param auto_generate_annotate_fields: 指定将自动生成的annotate的字段, 为[True, '__all__']时自动替换为base_fields
    :return:
    """
    base_model = get_base_model(base_model)
    field_names = get_field_names_by_model(base_model)
    if base_fields != '__all__' and auto_generate_annotate_fields is None:
        if len(set(base_fields) - set(field_names)):
            auto_generate_annotate_fields = True

    if auto_generate_annotate_fields is None:
        class BaseSerializer(s.ModelSerializer):
            class Meta:
                model = base_model
                fields = base_fields

        base_serializer = BaseSerializer
        return base_serializer
    else:
        if auto_generate_annotate_fields is True or auto_generate_annotate_fields in ['__all__', ['__all__']]:
            auto_generate_annotate_fields = base_fields
        elif isinstance(auto_generate_annotate_fields, list):
            auto_generate_annotate_fields = auto_generate_annotate_fields + field_names
        else:
            raise ValueError(f"auto_generate_annotate_fields[{auto_generate_annotate_fields}]取值错误!")

        # --- 这里要把queryset里有, 但model.fields里没有的字段在serializers时自动加上
        new_dc_ls = []
        new_func_ls = []
        for field in auto_generate_annotate_fields:
            if field not in field_names and field not in '__all__':
                # 指定方法字段 SerializerMethodField 后, 再增加 get_field_function.
                field_method = s.SerializerMethodField()
                new_func_name = f'get_{field}'

                def get_field_value(self, obj):
                    ret = None
                    k_name = getattr(self, '__field_name__')
                    if hasattr(obj, k_name):
                        ret = getattr(obj, k_name)
                    elif isinstance(obj, dict) and k_name in obj:
                        ret = obj.get(k_name)
                    elif '__' in k_name:
                        qs_ls = get_base_queryset(obj)
                        ret = qs_ls.filter(pk=obj.pk).values(k_name)[0].get(k_name)
                    else:
                        msg = f'自动生成的[{k_name}]字段值为空! --- from get_base_serializer'
                        warn(msg)
                    return ret

                new_func_cls = {
                    '__field_name__': field,
                    new_func_name: get_field_value,
                }
                new_func_cls = type("new_func_cls", (object,), new_func_cls)
                new_func = getattr(new_func_cls(), new_func_name)

                new_dc_i = {
                    field: field_method,
                }
                new_func_i = {
                    new_func_name: new_func,
                }
                new_dc_ls.append(new_dc_i)
                new_func_ls.append(new_func_i)

        # --- 生成新序列化器base_serializer
        meta_dc = {
            'model': base_model,
            'fields': base_fields
        }
        Meta = type("Meta", (object,), meta_dc)
        cls_dc = {
            'Meta': Meta,
        }
        for i in range(len(new_dc_ls)):
            cls_dc.update(new_dc_ls[i])
            cls_dc.update(new_func_ls[i])
        base_serializer = type("BaseSerializer", (s.ModelSerializer,), cls_dc)
        return base_serializer


def judge_is_obj_level_of_request(request):
    """
    判断本次访问是否为obj对象级, 否则就是model模型级
    """
    if 'pk' in request._request.resolver_match.kwargs:
        return True
    else:
        return False


def conv_queryset_ls_to_serialzer_ls(qs_ls: list):
    """
    qs_ls: 多个queryset数据的序列化, 手动转化为dc_ls
    """
    dc_ls = []
    if not qs_ls:
        return dc_ls

    q = qs_ls[0]
    if not isinstance(q, dict):
        q = model_to_dict(q)
    kname_ls = list(q.keys())

    dc_ls = []
    for q in qs_ls:
        # print(q)
        for kname_i in kname_ls:
            dc = {
                kname_i: getattr(q, kname_i)
            }
            dc_ls.append(dc)
    return dc_ls


def get_field_type_in_db(model, field_name):
    """根据模型和字段名, 获取这个字段在数据库中对应的类型"""
    tp = model._meta.get_field(field_name).get_internal_type()
    return tp


def convert_db_field_type_to_python_type(tp):
    tp = re.sub(r'\(.*\)', '', tp)      # 删除括号内的内容, 如"CharField(source='more_group.explain') "
    if tp in ['TextField', 'CharField', 'DateField', 'DateTimeField', 'SerializerMethodField']:
        field_type = 'str'
    elif tp in ['IntegerField', 'AutoField', 'BigAutoField']:
        field_type = 'int'
    elif tp in ['FloatField']:
        field_type = 'float'
    elif tp == 'BooleanField':
        field_type = 'bool'
    elif '=' in tp:
        # 类, 一般返回一个dc_ls类型
        field_type = 'list'
    # elif tp == 'FileField':
    #     field_type = 'file'
    else:
        field_type = tp
    return field_type


def get_field_type_in_py(model, field_name):
    """根据模型和字段名, 获取这个字段在python中对应的类型"""
    tp = get_field_type_in_db(model, field_name)
    field_type = convert_db_field_type_to_python_type(tp)
    return field_type


def reset_db_sequence(model):
    """重置数据库索引, 避免postgresql在手动导入csv/excel后出错."""
    md = get_base_model(model)
    sequence_sql = connection.ops.sequence_reset_sql(no_style(), [md])
    with connection.cursor() as cursor:
        for sql in sequence_sql:
            cursor.execute(sql)
    cursor.close()


def APIResponse(ret=None, status=200, msg=None):
    if isinstance(ret, Response):
        ret = ret.data
    ret = pure.add_status_and_msg(ret, status=status, msg=msg)
    ret = Response(ret)
    return ret


class Pagination(PageNumberPagination):
    """
    * 默认分页器参数设置

    - page_size: 每页16个
    - page_size_query_param: 前端控制每页数量时使用的参数名, 'page_size'
    - page_query_param: 页码控制参数名"p"
    - max_page_size: 最大1000页
    """
    page_size = int(PAGINATION_SETTINGS.get('page_size'))
    page_size_query_param = PAGINATION_SETTINGS.get('page_size_query_param', 'page_size')
    page_query_param = PAGINATION_SETTINGS.get('page_query_param', 'p')
    max_page_size = int(PAGINATION_SETTINGS.get('max_page_size'))


class StateMsgResultJSONRenderer(JSONRenderer):

    def render(self, data, accepted_media_type=None, renderer_context=None):
        if 'status' not in data and 'msg' not in data:
            if 'detail' in data:
                e = data.pop('detail')

                msg = str(e)
                if data:
                    msg += str(data)

                if isinstance(e, ErrorDetail) and e.code == 'permission_denied':
                    status = 403
                else:
                    print('! ************ 莫名返回格式 StateMsgResultJSONRenderer **************')
                    status = 404
                    try:
                        msg = f"[{str(e.code)}] {msg}"
                    except:
                        pass
                data = {
                    'status': status,
                    'msg': msg,
                    'result': [],
                }
            else:
                pass
        return super(StateMsgResultJSONRenderer, self).render(data, accepted_media_type, renderer_context)


def get_base_model(obj) -> Model:
    """判断是Queryset还是BaseModel"""
    if isinstance(obj, QuerySet):
        return obj.model
    else:
        if isinstance(obj, ContentType):
            base_model = obj.model_class()
            return base_model

        if hasattr(obj, 'objects'):
            # BaseModel
            return obj
        elif hasattr(obj.__class__, 'objects'):
            # 单个obj
            return obj.__class__
        else:
            return obj


def get_base_queryset(obj) -> QuerySet:
    """
    返回所有obj类型的QuerySet
    """
    ret = get_base_model(obj)
    if ret:
        ret = ret.objects.all()
    return ret


def conv_to_queryset(obj) -> QuerySet:
    """
    强制转换为QuerySet
    """
    if not isinstance(obj, QuerySet):
        ret = get_base_model(obj)
        ret = ret.objects.all()
    else:
        ret = obj
    return ret


def paginate_qsls_to_dcls(qsls, serializer, page: int, per_page=16, context=None):
    """
    * 手动分页函数

    - 指定模型的queryset_ls和serializer, 然后按给定的page和per_page参数获取分页后的数据
    """

    page_size = int(per_page)

    if page_size == 0:
        page_dc = {
            "count_items": qsls.count(),
            "total_pages": None,
            "page_size": page_size,
            "p": int(page)
        }
        return [], page_dc

    p = Paginator(qsls, per_page)
    page_dc = {
        "count_items": int(p.count),
        "total_pages": int(p.num_pages),
        "page_size": page_size,
        "p": int(page)
    }

    page_obj = p.get_page(page)

    # --- 处理单个Model和多个Model的情况
    if serializer.__class__.__name__ == 'function':
        try:
            dc_ls = serializer(page_obj, context=context)
        except Exception as e:
            raise Exception(f'--- paginate_qsls_to_dcls错误!!! 2022/2/25, error: {e}')
    else:
        dc_ls = serializer(page_obj, many=True, context=context).data
    return dc_ls, page_dc


def conv_queryset_to_dc_ls(queryset: QuerySet):
    dc_ls = []
    for q in queryset:
        dc_ls.append(q)
    return dc_ls


def order_qs_ls_by_id(qs_ls, sort_by='id', ascending=True):
    df = pd.DataFrame(qs_ls)
    if df.empty:
        return []
    df = df.sort_values(by=sort_by, ascending=ascending)

    cols = df.columns
    dc_ls = []
    for i, row in df.iterrows():
        dc = {}
        for j in range(len(cols)):
            _dc = {cols[j]: row.get(cols[j])}
            dc.update(_dc)
        dc_ls.append(dc)
    return dc_ls


def order_by_order_type_ls(queryset, order_type_ls) -> QuerySet:
    """
    根据传入的order_type_ls参数, 对queryset进行排序.

    若order_type_ls包含__None__, 则清空queryset的排序规则.
    """
    if '__None__' in order_type_ls:
        ret = queryset.order_by()
    elif order_type_ls is None or isinstance(order_type_ls, str):
        ret = order_by_order_type(queryset, order_type_ls)
    else:
        ls = []
        for order_type in order_type_ls:
            if order_type:
                if order_type.startswith('-'):
                    order_type1 = order_type[1:]
                    ls.append(F(order_type1).desc(nulls_last=True))
                else:
                    ls.append(F(order_type).asc(nulls_first=True))
        ret = queryset.order_by(*ls)
    return ret


def order_by_order_type(queryset, order_type=None):
    ret = queryset
    if order_type:
        if order_type.startswith('-'):
            order_type1 = order_type[1:]
            ret = queryset.order_by(F(order_type1).desc(nulls_last=True))
        else:
            ret = queryset.order_by(F(order_type).asc(nulls_first=True))

    return ret


def api_decorator(func):
    """
    * API装饰器

    - 如果运行出错, 将格式化输出错误的信息, 并返回给前端, 而不会报错.
    - 自动处理postgresql中idle状态connection过多的情况
    """
    @wraps(func)
    def wrapped_function(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            print('--- API Error! ---')
            print(e)
            msg = f'Error! {str(e)}'

            e_str = str(e)
            if 'client' in e_str:
                msg += '!!!!!! 可能出现postgresql的idle链接状况???'
                print(msg)
                # --- postgres的idle链接需要解决, 关闭旧链接(以下使用), 或单线程运行`manage.py runserver --nothreading`
                from django.db import close_old_connections
                from django.db import connection
                close_old_connections()
                with connection.cursor() as cursor:
                    sql = "SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE state = 'idle'"
                    cursor.execute(sql)
                    row = cursor.fetchall()
                    print(sql)
                    print(row)
            my_api_assert_function(False, msg=msg, status='404')
    return wrapped_function


def get_model_max_id_in_db(model):
    """
    仅适用于postgresql, mysql直接忽略就行.
    """
    meta = model._meta
    if not callable(model):
        model = type(model)
    ordering = meta.ordering

    assert sum([1 if f.name == 'id' and f.primary_key is True else 0 for f in meta.fields]), "Model的主键必须是id!"

    if ordering:
        if ordering[0] == '-id':
            obj = model.objects.first()
            max_id = obj.id if obj else 0
            ret = max_id + 1
            return ret
        if ordering[0] == 'id':
            obj = model.objects.last()
            max_id = obj.id if obj else 0
            ret = max_id + 1
            return ret
    qs = model.objects.all()
    max_id = qs.aggregate(max_id=Max('id')).get('max_id')  if qs.count() else 0
    ret = max_id + 1
    return ret


def old_get_model_max_id_in_db(model):
    meta = model._meta

    assert sum([1 if f.name == 'id' and f.primary_key is True else 0 for f in meta.fields]), "Model的主键必须是id!"

    cursor = connection.cursor()
    # db_prefix = meta.__str__().split('.')[0]

    # --- 先尝试创建id_seq
    id_seq = f"{meta.app_label}_{meta.db_table}_id_seq"

    try:
        sql = f"""CREATE SEQUENCE IF NOT EXISTS {id_seq}"""
        cursor.execute(sql)
    except Exception as e:
        # mysql不执行本函数也能正常运行
        print(e)
        print('不是PostgreSQL无法运行CREATE SEQUENCE语句! 请确认数据库类型!')

    # --- 找出最大的id
    sql = f"""select setval('{id_seq}', (select max(id) from "{meta.db_table}")+1);"""
    print('---', sql)
    # sql = '(select max(id) from  "{meta.db_table}")'
    # print('sql---', sql)
    cursor.execute(sql)
    row = cursor.fetchall()
    curr_id = row[0][0]
    ret = 0 if curr_id is None else curr_id
    cursor.close()
    return ret


def get_abs_order_type_ls(order_type_ls):
    if isinstance(order_type_ls, str):
        order_type_ls = [order_type_ls]
    ret = [re.sub(r'^-', '', field_name) for field_name in order_type_ls]
    return ret


def get_executable_sql(queryset):
    """
    输入queryset, 获得可直接执行的sql语句
    """
    cursor = connection.cursor()
    sql, params = queryset.query.sql_with_params()
    prefix = 'EXPLAIN '
    cursor.execute(prefix + sql, params)
    sql: str = cursor.db.ops.last_executed_query(cursor, sql, params)
    sql = sql[len(prefix):]
    cursor.close()
    return sql


def get_MySubQuery(my_model, field_name, function_name, output_field=m.IntegerField, alias=None):
    """
    # 获取子查询

    ## 简介
    - 主要用在进行`qs_ls.all().order_by().order_by(field_name).distinct(field_name)`后, 再进行`annotate`操作.
    - 普通的`m.SubQuery`操作将在`distinct`后的`annotate`中报错!

    ## 参考
    - [django文档_1](https://django-orm-cookbook-zh-cn.readthedocs.io/zh_CN/latest/subquery.html)
    - [django文档_2](https://docs.djangoproject.com/zh-hans/4.0/ref/models/expressions/)

    :param my_model: 指定模型, 用以获取模型基本属性`meta`. 若为空, 则认为是annotate字段, 使用默认的field_name.
    :param field_name: 用来进行计算的字段名
    :param function_name: 要在数据库中调用的函数名
    :param output_field: 使用Query计算后, 输出的字段类型
    :param alias: 计算后储存结果变量名, 默认为`tmp`
    :return: 子查询类`MySubQuery`

    ## eg:
    SQCount = get_MySubQuery(my_model=classification_qs_ls_0, field_name=foreign_key_name, function_name='Count', output_field=m.IntegerField)
    classification_qs_ls = classification_qs_ls_0.annotate(
        # 每个学科有多少种
        count=SQCount(
            classification_qs_ls_0.filter(**{key: m.OuterRef(key)})
        ),
    )
    """
    base_model = get_base_model(my_model)
    meta = base_model._meta

    field_names = [field.name for field in meta.fields]
    db_column_names = [field.db_column if field.db_column else field.name for field in meta.fields]
    field_dc = dict(zip(field_names, db_column_names))
    db_column_name = field_dc.get(field_name)  # 获取字段在db中的列名
    db_column_name = db_column_name if db_column_name else field_name       # 没有的话, 就用默认field_name

    alias = 'tmp' if not alias else alias
    my_template = f"(SELECT {function_name}({db_column_name}) FROM (%(subquery)s) {alias})"
    my_output_field = output_field

    class MySubQuery(m.Subquery):
        template = my_template
        output_field = my_output_field() if isinstance(my_output_field, type) else my_output_field

    return MySubQuery


def get_obj_by_content_type(obj_id, model_name, app_label):
    ct_qs_ls = ContentType.objects.filter(app_label=app_label, model=model_name)
    assert ct_qs_ls.count() == 1, f'ContentType数量不为1! current_value: {ct_qs_ls.count()}'
    ct_qs_i = ct_qs_ls[0]
    base_model = ct_qs_i.model_class()
    obj = base_model.objects.get(id=obj_id)
    return obj


def get_QS_by_dc(dc, add_type):
    """
    根据dc返回QS
    :param dc: 过滤条件
    :param add_type: 合并逻辑
    :return: QS
    """
    QS = Q()
    for k, v in dc.items():
        d = {k: v}
        QS.add(Q(**d), add_type)
    return QS


def get_model_verbose_name_dc():
    """
    获得model_verbose_name对应的ContentType的id
    """
    ct_qs_ls = ContentType.objects.all()
    dc = {}
    for ct_qs_i in ct_qs_ls:
        base_model = ct_qs_i.model_class()
        if base_model is not None:
            k = base_model._meta.verbose_name
            # v = ct_qs_i.model
            v = ct_qs_i
            dc.update({k: v})
    return dc


def get_user_ip(request):
    context = request.parser_context
    if 'HTTP_X_FORWARDED_FOR' in context["request"].META:
        user_ip = context["request"].META['HTTP_X_FORWARDED_FOR']
    else:
        user_ip = context["request"].META['REMOTE_ADDR']
    return user_ip


def update_none_to_zero_by_field_name(qs_ls, field_name):
    """
    将qs_ls中field_name字段的None改为0
    """
    filter_dc = {
        f'{field_name}__isnull': True
    }
    update_dc = {
        field_name: 0
    }
    qs_ls.filter(**filter_dc).update(**update_dc)


def get_df_by_freq_and_year(
        queryset,
        frequency_cname=None,
        aggregate_method_name='Sum',
        output_col_name=None,
        year_field_name='year',
        complete_year_ls=False,
        year_range_ls=None,
):
    """
    获取加权的年份分布图

    * 这个比较快

    - aggregate_method_name, frequency_cname: 用来aggregate的方法和字段
    - complete_year_ls: 补全中间年份
    - output_col_name: 输出列名
    - year_field_name: 年份字段名
    """
    year_qsv_ls = queryset.values(year_field_name).distinct(year_field_name).order_by(year_field_name)
    year_ls = [dc.get(year_field_name) for dc in year_qsv_ls]
    assert hasattr(m, aggregate_method_name), f'django.db.models不存在[{aggregate_method_name}]方法!'
    aggregate_method = getattr(m, aggregate_method_name)

    output_col_name = output_col_name if output_col_name  else frequency_cname

    if complete_year_ls:
        if year_range_ls:
            year_min, year_max = year_range_ls
        else:
            year_min, year_max = min(year_ls), max(year_ls)
        year_range = list(range(year_min, year_max + 1))
    else:
        year_range = year_ls

    year_distribution_dc_ls = []
    for year in year_range:
        aggregate_dc = {'tmp': aggregate_method(frequency_cname)}
        value = queryset.filter(year=year).aggregate(**aggregate_dc)
        value = 0 if value is None or value.get('tmp') is None else value.get('tmp')

        year_distribution_dc = {
            year_field_name: year,
            output_col_name: value,
        }
        year_distribution_dc_ls.append(year_distribution_dc)
    return year_distribution_dc_ls


def judge_db_is_migrating():
    if 'makemigrations' in sys.argv or 'migrate' in sys.argv:
        return True
    else:
        return False


def get_total_occurrence_times_by_keywords(total_qs_ls, search_field_ls, keywords, get_frequence=False):
    """
    统计keywords在qs_ls的search_field_ls中是否出现

    :param total_qs_ls: 用来统计的queryset
    :param search_field_ls: 要匹配的字段
    :param kw: 用来检索的关键词
    :param get_frequence: 是否精确计算kw在字段中出现的次数
    :return: queryset, 且annotate出现次数, 存在`total_occurrence_times`字段中
    """
    from django.db.models import functions

    my_api_assert_function(isinstance(keywords, list), 'keywords的类型必须为list!')
    search_dc = {}
    occurrence_times_ls = []

    for k in range(len(keywords)):
        kw = keywords[k]
        kw_l = len(kw)
        for sf_i in search_field_ls:
            if get_frequence:
                k_name = f'k{k}_{sf_i}_occurrence_times'
                dc = {
                    k_name: (functions.Length(sf_i) - functions.Length(
                        functions.Replace(sf_i, m.Value(kw), m.Value('')))) / kw_l  # 出现次数
                }
            else:
                k_name = f'k{k}_in_{sf_i}'
                dc = {
                    k_name: m.Exists(total_qs_ls.filter(id=m.OuterRef('id')).filter(**{f'{sf_i}__contains': kw})),   # 判断是否在title中
                }
            occurrence_times_ls.append(k_name)
            search_dc.update(dc)
    res_qs_ls: m.QuerySet = total_qs_ls.annotate(**search_dc)
    # show_ls(res_qs_ls.values(*(['id'] + search_field_ls + list(search_dc.keys())))[:3])

    f_ls = 0
    for i in occurrence_times_ls:
        if get_frequence:
            f_ls += F(i)
        else:
            f_ls += m.Case(
                m.When(**{i: True}, then=m.Value(1)),
                default=0,
                output_field=m.IntegerField()
            )

    res_qs_ls = res_qs_ls.annotate(**{'total_occurrence_times': f_ls})
    ret = res_qs_ls
    return ret





#!/usr/bin/python
# -*- coding: UTF-8 -*-

import os
import zipfile
import re


def zipDir(dirpath, outFullName, remove_root_dir_path=None):
    """
    压缩指定文件夹
    :param dirpath: 目标文件夹路径
    :param outFullName: 压缩文件保存路径+xxxx.zip
    :param remove_root_dir_path: 是否去掉目标根路径. 若为str, 则删除该根路径文件名, 若为1, 则删除所有根路径. 2则保留最后一层
    :return: 无
    """

    if remove_root_dir_path is None:
        # 保留目标根路径
        with zipfile.ZipFile(outFullName, 'w') as target:
            for i in os.walk(dirpath):
                for n in i[2]:
                    target.write(''.join((i[0], '\\', n)))

    elif isinstance(remove_root_dir_path, str):
        with zipfile.ZipFile(outFullName, "w", zipfile.ZIP_DEFLATED) as zip:
            for path, dirnames, filenames in os.walk(dirpath):
                # path = path.replace('/', '\\')
                path = path.replace('/', os.sep)
                remove_root_dir_path = remove_root_dir_path.replace('/', os.sep)
                if remove_root_dir_path.endswith('\\') or remove_root_dir_path.endswith('/'):
                    remove_root_dir_path = remove_root_dir_path[:-1]

                # 去掉目标根路径，只对目标文件夹下边的文件及文件夹进行压缩
                # fpath = path.replace(dirpath, '')
                pattern = fr"""^{remove_root_dir_path.encode('unicode_escape').decode('ascii')}"""
                fpath = re.sub(pattern, '', path)
                # print(path, '---', fpath, filenames)

                for filename in filenames:
                    # if fpath.startswith(remove_root_dir_path):
                    #     # 大坑, 注意re使用fr-string需要重新编码!
                    #     pattern = fr"""^{remove_root_dir_path.encode('unicode_escape').decode('ascii')}"""
                    #     # pattern = fr"""^{remove_root_dir_path}"""     # 不行
                    #     fpath = re.sub(pattern, '', fpath)
                    #     # fpath = fpath[len(remove_root_dir_path):]
                    # fpath = re.sub(r'^' + remove_root_dir_path, '', fpath)
                    # print(os.path.join(path, filename), '------', os.path.join(fpath, filename))
                    zip.write(os.path.join(path, filename), os.path.join(fpath, filename))

    elif remove_root_dir_path == 1:
        with zipfile.ZipFile(outFullName, "w", zipfile.ZIP_DEFLATED) as zip:
            for path, dirnames, filenames in os.walk(dirpath):
                fpath = path.replace(dirpath, '')
                print(fpath, path, dirnames, filenames)
                # 去掉目标根路径，只对目标文件夹下边的文件及文件夹进行压缩
                for filename in filenames:
                    zip.write(os.path.join(path, filename), os.path.join(fpath, filename))
    elif remove_root_dir_path == 2:     # 保留最后一层文件夹模式
        with zipfile.ZipFile(outFullName, "w", zipfile.ZIP_DEFLATED) as zip:
            base_dirpath = os.path.basename(dirpath)
            for path, dirnames, filenames in os.walk(dirpath):
                fpath = path.replace(dirpath, base_dirpath)
                # 去掉目标根路径，只对目标文件夹下边的文件及文件夹进行压缩
                for filename in filenames:
                    zip.write(os.path.join(path, filename), os.path.join(fpath, filename))


if __name__ == "__main__":
    import sys
    assert len(sys.argv) == 2, '必须将文件夹拖拽进来!'
    # input_path = "temp"
    input_path = sys.argv[1]
    print('input_path --- ', input_path)
    output_path = f"./{os.path.basename(input_path)}.zip"

    # zipDir(input_path, output_path)
    zipDir(input_path, output_path, remove_root_dir_path=2)



from .utils import *
from rest_framework.views import APIView


class BaseListView(ListModelMixin, RetrieveModelMixin, GenericAPIView):
    """
    * API: BaseModel的ListView和RetrieveView接口

    - 测试接口:
        - List:
            GET /api/index/BaseList/?order_type=-id&page_size=4&p=1
        - Retrieve:
            GET /api/index/BaseList/5/
    """

    _name = 'BaseListView'  # 这个在自动生成wiki时要用到

    # renderer_classes = (StateMsgResultJSONRenderer,)
    # renderer_classes = APIView.renderer_classes + [StateMsgResultJSONRenderer]

    pagination_class = Pagination

    filter_fields = []  # ['__all__']过滤所有. 精确检索的过滤字段, 如果过滤条件复杂的话, 建议重写

    order_type_ls = []
    distinct_field_ls = []
    only_get_distinct_field = False  # 仅返回distinct指定的字段

    serializer_class = None
    auto_generate_serializer_class = False
    base_fields = '__all__'  # 当auto_generate_serializer_class为True时, 将自动生成序列化器, 然后根据base_fields返回字段

    list_fields = None
    retrieve_fields = None
    retrieve_serializer_class = None  # retrieve对应的serializer_class
    list_serializer_class = None  # list对应的serializer_class

    retrieve_filter_field = 'pk'  # 详情页的查询字段名, 默认为 {{url}}/app/view/pk , 可由前端指定.

    method = None  # 中间变量, 记录请求是什么类型, list/retrieve等
    _tmp = None  # 无意义, 用来处理数据

    _post_type = None  # 用来判断是什么请求类型, 然后判断request_data取哪个
    post_type_ls = ["list", "retrieve", "bulk_list"]  # post请求方法
    create_unique = True  # 创建时是否允许重复

    convert_to_bool_flag = 'bool__'                       # 将特定格式强制转换为布尔变量, 如 `名字不为空: name__isnull=bool_0`
    negative_flag = '!'                                   # filter_fields 条件取否时使用, 如: `id不等于1: !id=1`

    # add_host_prefix_to_media_url = True                                    # 是否返回文件的时候加上当前域名的prefix

    def __new__(cls, *args, **kwargs):
        ret = super().__new__(cls, *args, **kwargs)

        if cls.list_serializer_class is None and cls.list_fields is not None:
            cls.list_serializer_class = get_base_serializer(cls.queryset, base_fields=cls.list_fields)

        if cls.retrieve_serializer_class is None and cls.retrieve_fields is not None:
            if cls.retrieve_fields == '__all__':
                cls.retrieve_serializer_class = get_base_serializer(
                    cls.queryset,
                    base_fields=cls.retrieve_fields,
                    auto_generate_annotate_fields=cls.list_fields
                )
            else:
                cls.retrieve_serializer_class = get_base_serializer(
                    cls.queryset,
                    base_fields=cls.retrieve_fields,
                )

        return ret

    def get(self, request, *args, **kwargs):
        """
        - 如果request携带pk参数, 则进行Retrieve操作, 否则进行List操作.

        - BaseList的默认get请求参数(仅在List操作时生效)
            - page_size: 分页器每页数量, 前端用来控制数据每页展示的数量, 在Pagination类中设置.
            - p: 第p页.
            - order_type_ls: 排序字段, 如"id"和"-id".
        """
        if kwargs.get('pk'):
            self.method = 'retrieve'
            ret, status, msg = self.get_retrieve_ret(request, *args, **kwargs)
        else:
            self.method = 'list'
            ret, status, msg = self.get_list_ret(request, *args, **kwargs)
        return APIResponse(ret, status=status, msg=msg)

    def bulk_list(self, request, *args, **kwargs):
        """
        根据id批量删除
        """
        query_dc = self.get_request_data()

        id_ls = get_list(query_dc, 'id_ls')
        self.method = query_dc.get('http_method', 'retrieve')  # 优先返回retrieve详情数据

        my_api_assert_function(id_ls, msg=f'id_ls[{id_ls}]不能为空!!!')
        my_api_assert_function(isinstance(id_ls, list),
                               msg=f'id_ls[{id_ls}]应为list类型, 不应为{id_ls.__class__.__name__}类型!!')

        base_model = get_base_model(self.queryset)
        qs_ls = base_model.objects.filter(id__in=id_ls)
        self.queryset = qs_ls

        qs_ls = self.get_ordered_queryset()

        page_size = query_dc.get('page_size', self.pagination_class.page_size)
        p = query_dc.get('p', 1)
        data, page_dc = paginate_qsls_to_dcls(qs_ls, self.get_serializer_class(), page=p, per_page=page_size)
        ret = {
            'page_dc': page_dc,
            'data': data
        }
        return APIResponse(ret)

    def post(self, request, *args, **kwargs):
        """用post方法来跳转"""
        post_type = request.data.get('post_type', 'list')
        self._post_type = post_type
        my_api_assert_function(not post_type or post_type in self.post_type_ls,
                               f"操作类型post_type指定错误! 取值范围: {self.post_type_ls}")

        if post_type in ['list', 'retrieve']:
            return self.get(request, *args, **kwargs)
        elif post_type == 'bulk_list':
            return self.bulk_list(request, *args, **kwargs)
        else:
            return APIResponse(None, status=404, msg=f'请指定post操作类型, 取值范围: {self.post_type_ls}?')

    def get_serializer_class(self):
        request_data = self.get_request_data()

        if self.method == 'retrieve':
            ret = self.retrieve_serializer_class or self.serializer_class
        elif self.method == 'list':
            ret = self.list_serializer_class or self.serializer_class
        else:
            ret = self.retrieve_serializer_class or self.serializer_class or self.list_serializer_class

        if (self.auto_generate_serializer_class and ret is None) or request_data.get('base_fields'):
            ret = get_base_serializer(self.queryset, base_fields=self.get_base_fields())

        # --- 仅获取distinct之后的字段
        only_get_distinct_field = self.get_only_get_distinct_field()
        if only_get_distinct_field:
            distinct_field_ls = self.get_distinct_field_ls()
            assert distinct_field_ls, '指定了only_get_distinct_field的同时必须指定distinct_field_ls!'

            if isinstance(distinct_field_ls, str):
                distinct_field_ls = [distinct_field_ls]

            ret = get_base_serializer(self.queryset, distinct_field_ls)
            return ret

        assert ret, '返回的serializer_class不能为空!'
        return ret

    def get_only_get_distinct_field(self):
        key = 'only_get_distinct_field'
        value = self.get_key_from_query_dc_or_self(key)
        ret = pure.convert_query_parameter_to_bool(value)
        return ret

    def get_distinct_field_ls(self):
        key = 'distinct_field_ls'
        ret = self.get_key_from_query_dc_or_self(key, get_type='list')
        return ret

    def get_base_fields(self):
        key = 'base_fields'
        ret = self.get_key_from_query_dc_or_self(key, get_type='list')
        return ret

    def get_retrieve_ret(self, request, *args, **kwargs):
        """
        Retrieve操作

        - pk必须在`url.py::urlpatterns`中设置, 如: path('BaseList/<str:pk>/', views.BaseList.as_view())
        """
        status = 200
        msg = 'ok'

        try:
            ret = self.retrieve(request)
        except Exception as e:
            # # 这里有坑, 因为有可能返回的是已经自定义好了的APIException, 例如权限不足错误.
            # if not isinstance(e, APIException):
            #     # 没找到的情况, 404 Not Found
            #     ret = None
            #     status = 404
            #     msg = str(e)
            #     my_api_assert_function(ret, msg, status)
            # else:
            raise e

        return ret, status, msg

    def get_object(self):
        """retrieve时, object的获取"""
        retrieve_filter_field = self.get_key_from_query_dc_or_self('retrieve_filter_field')

        pk = self.request.parser_context.get('kwargs').get('pk')
        dc = {retrieve_filter_field: pk}
        queryset = self.filter_queryset(self.get_queryset())
        # queryset = queryset.filter(Q(**dc))
        if isinstance(queryset, QuerySet):
            queryset = queryset.filter(Q(**dc))
        else:
            queryset = queryset.objects.filter(Q(**dc))
        # print(get_executable_sql(queryset))

        count = queryset.count()
        if count != 1:
            if count > 1:
                my_api_assert_function(False, f'检索结果不唯一!')
            else:
                my_api_assert_function(False, f'未检索到{dc}对应的内容!')

        obj = queryset[0]
        self.check_object_permissions(self.request, obj)
        return obj

    def get_list_ret(self, request, *args, **kwargs):
        """
        List操作
        """
        status = 200
        msg = 'ok'

        # --- 得到list方法的queryset
        self.queryset = self.get_list_queryset()

        if not isinstance(self.queryset, QuerySet):
            self.queryset = self.queryset.objects.all()

        # --- 根据ordering参数, 获得排序后的queryset
        self.queryset = self.get_ordered_queryset()

        # --- 获取返回数据(转化为特定格式)
        try:
            ret = self.list(request)
        except Exception as e:
            # 页码无效的情况, 404 Not Found
            raise e
            # ret = None
            # my_api_assert_function(ret, f'List error! --- {e}', 404)
            # status = 404
            # msg = str(e)
        return ret, status, msg

    def get_request_data(self) -> dict:
        """
        请求所携带的数据, 除了get方法跳过来的外, 均以请求体body携带的数据request.data优先.
        """
        if not hasattr(self, 'request'):
            return {}

        if self._post_type is None:
            if self.request.method and self.request.method not in ['GET', 'HEAD', 'OPTIONS']:
                ret = self.request.data
            else:
                ret = self.request.query_params
        else:
            # ret = self.request.GET or self.request.query_params or self.request.data
            ret = self.request.data or self.request.query_params

        if hasattr(self, '_set_request_data') and getattr(self, '_set_request_data'):
            ret = self._set_request_data
        return ret

    def set_request_data(self, request_data):
        self._set_request_data = request_data

    def get_ordered_queryset(self):
        """
        按order_type_ls指定的字段排序
        """
        query_dc = self.get_request_data()
        # order_type_ls = get_list(query_dc, key='order_type_ls') \
        #     if get_list(query_dc, key='order_type_ls') else self.order_type_ls

        # distinct_field_ls = get_list(query_dc, key='distinct_field_ls')\
        #     if get_list(query_dc, key='distinct_field_ls') else self.distinct_field_ls

        order_type_ls = self.get_key_from_query_dc_or_self('order_type_ls', get_type='list')
        distinct_field_ls = self.get_key_from_query_dc_or_self('distinct_field_ls', get_type='list')

        if not order_type_ls:
            # 旧版本可能用的order_type, 尝试赋值
            order_type = get_list(query_dc, key='order_type')
            if order_type:
                order_type_ls = order_type

        qs_ls = self.queryset

        # distinct操作
        if distinct_field_ls and distinct_field_ls not in ['__None__', ['__None__']]:
            assert isinstance(distinct_field_ls, (list, tuple)), 'distinct_field_ls因为list或者tuple!'
            qs_ls = qs_ls.order_by(*distinct_field_ls).distinct(*distinct_field_ls)  # bug: distinct_field_ls 后的字段无法排序

        # order_by操作
        if pure.convert_query_parameter_to_bool(order_type_ls):
            if distinct_field_ls and distinct_field_ls not in ['__None__', ['__None__']]:
                qs_ls = self.queryset.filter(pk__in=qs_ls.values('pk'))

            try:
                qs_ls = order_by_order_type_ls(qs_ls, order_type_ls)
            except ValueError as e:
                msg = f'参数order_type_ls指定的排序字段[{order_type_ls}]排序失败! 更多信息: {str(e)}'
                raise ValueError(msg)
        else:
            qs_ls = order_by_order_type_ls(qs_ls, ['pk'])       # 默认按照pk排序

        self.queryset = qs_ls
        return self.queryset

    def get_list_queryset(self):
        """
        得到queryset, 仅对list方法生效
        """
        self.queryset = self.get_queryset()
        if not isinstance(self.queryset, QuerySet):
            self.queryset = self.queryset.objects.all()
        self.run_list_filter()
        return self.queryset

    def run_list_filter(self):
        """
        返回用self.filter_fields过滤后的queryset列表
        """
        if self.queryset.exists():
            """
            过滤字段filter_fields
            """
            query_dc = self.get_request_data()
            FILTER_ALL_FIELDS = True if self.filter_fields in ['__all__', ['__all__']] else False

            self.queryset = self.get_queryset()
            base_model = get_base_model(self.queryset)
            if not base_model.objects.exists():     # exists省性能
                return self.queryset

            meta = base_model.objects.first()._meta
            field_names = [field.name for field in meta.fields]
            many_to_many_field_names = [field.name for field in meta.many_to_many]
            if many_to_many_field_names:
                field_names.extend(many_to_many_field_names)

            if self.queryset.exists():      # exists 比 count 省性能
                qs_i = self.queryset[0]
                if isinstance(qs_i, dict):
                    new_names = list(qs_i.keys())
                else:
                    new_dc: dict = qs_i.__dict__.copy()
                    new_dc.pop('_state')
                    new_names = list(new_dc.keys())

                # 可能用annotate增加了注释字段, 所以要处理一下, 避免过滤出错
                for new_name in new_names:
                    if new_name not in field_names:
                        field_names.append(new_name)

            field_names.append('pk')  # 将主键pk加进去作为过滤条件

            for fn, value in query_dc.items():
                filter_flag = False     # 是否过滤
                negative_flag = False   # 是否取否

                if fn:
                    # fn: str
                    if fn.startswith(self.negative_flag):
                        fn = fn[1:]
                        negative_flag = True

                    if f'__' in fn:
                        _fn = fn.split('__')[0]
                        if _fn in field_names or fn in field_names:
                            if FILTER_ALL_FIELDS or fn in self.filter_fields or _fn in self.filter_fields or fn == 'pk' or _fn == 'pk':
                                filter_flag = True
                    else:
                        if fn in field_names and (FILTER_ALL_FIELDS or fn in self.filter_fields or fn == 'pk'):
                            filter_flag = True

                if filter_flag:
                    if value is not None and value != '':  # 默认为空字符串时, 将不作为过滤条件
                        # print(fn, value)
                        convert_to_bool_flag = self.convert_to_bool_flag
                        if isinstance(value, str) and value.startswith(convert_to_bool_flag):
                            value = pure.convert_query_parameter_to_bool(value[len(convert_to_bool_flag):])

                        if isinstance(value, str) and fn.endswith('__isnull'):
                            value = pure.convert_query_parameter_to_bool(value)

                        dc = {fn: value}
                        if negative_flag:
                            self.queryset = self.queryset.exclude(**dc)
                        else:
                            self.queryset = self.queryset.filter(**dc)

        return self.queryset

    def _get_list_queryset(self):  # 兼容问题
        return self.get_list_queryset()

    def list(self, request, *args, **kwargs):
        query_dc = self.get_request_data()
        page_size = query_dc.get('page_size', self.pagination_class.page_size)

        GET_ALL = -999      # 获取全部
        p = query_dc.get('p', 1)

        try:
            p = int(p)

            if p == GET_ALL or int(page_size) == GET_ALL:
                p = 1
                page_size = 999999999

            my_api_assert_function(p > 0, f'页码p[{p}]必须为[正整数]!')
        except Exception as e:
            my_api_assert_function(0, f'页码[p: {p}] 必须为[正整数]!')

        context = self.get_serializer_context()

        serializer_class = self.get_serializer_class()
        resp, page_dc = paginate_qsls_to_dcls(self.queryset, serializer_class, page=p, per_page=page_size,
                                              context=context)
        ret = self._conv_data_format(resp, page_dc)
        return ret

    def conv_data_format(self, data: (dict, Response), page_dc):
        ret = {
            'page_dc': page_dc,
            'data': data,
        }
        return ret

    def _conv_data_format(self, *args, **kwargs):
        return self.conv_data_format(*args, **kwargs)

    def get_key_from_query_dc_or_self(self, key, get_type=None):
        """
        优先检索query_dc是否有key, 其次检索self是否有key这个属性
        :param key: 变量名
        :return:
        """

        query_dc = self.get_request_data()
        ret_0 = getattr(self, key) if hasattr(self, key) else None

        if get_type == 'list':
            data = get_list(query_dc, key)
            # if data == ['']:
            #     data = []
        elif get_type == 'bool':
            value = query_dc.get(key)
            if value is not None:
                data = pure.convert_query_parameter_to_bool(value)
                if data is False:
                    return data
            else:
                return ret_0
        else:
            data = query_dc.get(key)

        # 让request携带的数据可以覆盖自身的key值
        ret_1 = data if data else None
        ret = ret_1 or ret_0
        return ret

    def _get_key_from_query_dc_or_self(self, *args, **kwargs):
        return self.get_key_from_query_dc_or_self(*args, **kwargs)


class BaseList(BaseListView):  # 向下兼容, 返回格式调整, 重写_conv_data_format.
    def _conv_data_format(self, data: (dict, Response)):
        if isinstance(data, Response):
            data = data.data

            # 分页信息
        count = data.get('count')
        page_size = self.request.query_params.get('page_size', self.pagination_class.page_size)
        p = self.request.query_params.get('p', 1)
        total = math.ceil(count / int(page_size))
        page_dc = {
            'count': count,
            'total': total,
            'page_size': page_size,
            'p': p,
        }

        results = data.get('results')

        ret = {
            'page_dc': page_dc,
            'results': results,
        }
        return ret


from .mixins import MyCreateModelMixin, MyUpdateModelMixin, MyDestroyModelMixin


class CompleteModelView(BaseListView, MyCreateModelMixin, MyUpdateModelMixin, MyDestroyModelMixin):
    """
    * 一个模型的增删改查全套接口

    - 可在post方法中使用post_type指定操作类型.

    > 以下方法分别对应同一个url下的: 增, 删, 改, 查.

    - POST
        - 创建新数据
    - DELETE
        - 删除数据, 需指定id
    - PUT
        - 修改数据, 需指定id
    - GET
        - 查询列表页`GET url/`
        - 查询详情页, 需指定id.
          - 如: `GET url/id/`
    """
    _name = 'CompleteModelView'

    post_type_ls = ["list", "retrieve", "create", "update", "delete", "bulk_delete", "bulk_update", "bulk_list"]       # post请求方法
    _post_type = None
    create_unique = True        # 创建时是否允许重复

    def get_post_type(self):
        post_type = self.request.data.get('post_type', 'create')
        self._post_type = post_type
        my_api_assert_function(not post_type or post_type in self.post_type_ls, f"操作类型post_type指定错误! 取值范围: {self.post_type_ls}")
        return post_type

    def post(self, request, *args, **kwargs):
        """增"""
        post_type = self.get_post_type()
        # post_type = request.data.get('post_type', 'create')
        # self._post_type = post_type
        # my_api_assert_function(not post_type or post_type in self.post_type_ls, f"操作类型post_type指定错误! 取值范围: {self.post_type_ls}")

        if post_type == 'create':
            return self.create(request, *args, **kwargs)
        elif post_type == 'delete':
            return self.destroy(request, *args, **kwargs)
        elif post_type in ['list', 'retrieve']:
            return self.get(request, *args, **kwargs)
        elif post_type == 'update':
            return self.put(request, *args, **kwargs)
        elif post_type == 'bulk_delete':
            return self.bulk_delete(request, *args, **kwargs)
        elif post_type == 'bulk_update':
            return self.bulk_update(request, *args, **kwargs)
        elif post_type == 'bulk_list':
            return self.bulk_list(request, *args, **kwargs)
        else:
            return APIResponse(None, status=404, msg='请指定post操作类型, [create, update, delete]?')

    def delete(self, request, *args, **kwargs):
        """删"""
        return self.destroy(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        """改"""
        ret = self.partial_update(request, *args, **kwargs)
        return ret

    def bulk_delete(self, request, *args, **kwargs):
        """
        根据id批量删除
        """
        request_data = self.get_request_data()

        id_ls = get_list(request_data, 'id_ls')
        my_api_assert_function(id_ls, msg=f'id_ls[{id_ls}]不能为空!!!')
        my_api_assert_function(isinstance(id_ls, list), msg=f'id_ls[{id_ls}]应为list类型, 不应为{id_ls.__class__.__name__}类型!!')

        base_model = get_base_model(self.queryset)
        qs_ls = base_model.objects.filter(id__in=id_ls)
        qs_ls.delete()
        return APIResponse()

    def bulk_update(self, request, *args, **kwargs):
        """
        批量更新
        """
        request_data = self.get_request_data()

        id_ls = get_list(request_data, 'id_ls')
        field_dc = request_data.get('field_dc')

        my_api_assert_function(id_ls, msg=f'id_ls[{id_ls}]不能为空!!!')
        my_api_assert_function(isinstance(id_ls, list), msg=f'id_ls[{id_ls}]应为list类型, 不应为{id_ls.__class__.__name__}类型!!')
        my_api_assert_function(field_dc, 'field_dc不能为空!')

        if isinstance(field_dc, str):
            field_dc = json.loads(field_dc)

        # 开始批量更新, 注意只支持2层嵌套.
        foreign_key_field_dc = {}
        original_field_dc = {}
        for k, v in field_dc.items():
            dc = {k: v}
            if '__' in k:
                foreign_key_field_dc.update(dc)
            else:
                original_field_dc.update(dc)

        base_model = get_base_model(self.queryset)
        qs_ls: m.QuerySet = base_model.objects.filter(id__in=id_ls)
        my_api_assert_function(qs_ls.exists(), '未找到id_ls对应的数据!')

        qs_i = qs_ls[0]     # 样例数据

        # 先更新原生字段
        qs_ls.update(**original_field_dc)

        # 再依次更新外键字段
        for k, v in foreign_key_field_dc.items():
            fk_model_name, fk_field = k.split('__')
            dc = {fk_field: v}
            foreign_key_obj_i = getattr(qs_i, fk_model_name)
            foreign_key_model = get_base_model(foreign_key_obj_i)
            base_field = getattr(base_model, fk_model_name)
            assert hasattr(base_field, 'related'), '字段不是外键?'
            related = getattr(base_field, 'related')
            foreign_key_field_name = related.field.name
            filter_dc = {
                f'{foreign_key_field_name}__id__in': id_ls
            }
            qs_ls = foreign_key_model.objects.filter(**filter_dc)
            qs_ls.update(**dc)

        # 返回更新后的数据
        self.queryset = base_model.objects.filter(id__in=id_ls)
        # ret = self.get_serializer_class()(qs_ls, many=True).data
        ret, status, msg = self.get_list_ret(request, *args, **kwargs)
        return APIResponse(ret, status=status, msg=msg)


class DecoratorBaseListView(BaseListView):
    @api_decorator
    def get(self, request, *args, **kwargs):
        return super().get(*args, **kwargs)

    @api_decorator
    def post(self, request, *args, **kwargs):
        return super().post(*args, **kwargs)


class DecoratorCompleteModelView(CompleteModelView):
    """
    全部用api_decorator装饰
    """

    @api_decorator
    def get(self, request, *args, **kwargs):
        return super().get(*args, **kwargs)

    @api_decorator
    def post(self, *args, **kwargs):
        return super().post(*args, **kwargs)

    @api_decorator
    def put(self, request, *args, **kwargs):
        return super().put(*args, **kwargs)

    @api_decorator
    def delete(self, request, *args, **kwargs):
        return super().delete(*args, **kwargs)





import datetime
import xlwt
from xlwt import *


# 写入excel文件函数
def write2excel(n, head_data, records, download_url):
    """
    生成excel文件函数
    :param n:数据行数
    :param head_data:文件表头数据
    :param records:中间数据[[],[]]
    :param download_url:文件保存绝对路径
    :return:
    """
    # 获取时间戳
    time_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    # 工作表
    wbk = xlwt.Workbook()
    sheet1 = wbk.add_sheet('sheet1', cell_overwrite_ok=True)

    try:
        # 写入表头
        for filed in range(0, len(head_data)):
            sheet1.write(0, filed, head_data[filed], excel_head_style())

        # 写入数据记录
        for row in range(1, n+1):
            for col in range(0, len(head_data)):
                sheet1.write(row, col, records[row-1][col], excel_record_style())
                # 设置默认单元格宽度
                sheet1.col(col).width = 256*35
    except Exception as e:
        # logger.error(e)
        print(e)
    # print(type(time_str))
    download_url_fin = '{}_{}.xls'.format(download_url, time_str)
    try:
        wbk.save(download_url_fin)
    except Exception as e:
        print(e)
    return download_url_fin


# 定义文件表头格式
def excel_head_style():
    # 创建一个样式
    style = XFStyle()
    # 设置背景色
    pattern = Pattern()
    pattern.pattern = Pattern.SOLID_PATTERN
    # pattern.pattern_fore_colour = Style.colour_map['blue']  # 设置单元格背景色
    style.pattern = pattern
    # 设置字体
    font0 = xlwt.Font()
    font0.name = u'微软雅黑'
    font0.bold = True
    font0.colour_index = 0
    font0.height = 260
    style.font = font0
    # 设置文字位置
    alignment = xlwt.Alignment()  # 设置字体在单元格的位置
    alignment.horz = xlwt.Alignment.HORZ_CENTER  # 水平方向
    alignment.vert = xlwt.Alignment.VERT_CENTER  # 竖直方向
    style.alignment = alignment
    # 设置边框
    borders = xlwt.Borders()  # Create borders
    borders.left = xlwt.Borders.THIN  # 添加边框-虚线边框
    borders.right = xlwt.Borders.THIN  # 添加边框-虚线边框
    borders.top = xlwt.Borders.THIN  # 添加边框-虚线边框
    borders.bottom = xlwt.Borders.THIN  # 添加边框-虚线边框
    style.borders = borders

    return style


# 定义文件记录格式
def excel_record_style():
    # 创建一个样式
    style = XFStyle()
    # 设置字体
    font0 = xlwt.Font()
    font0.name = u'微软雅黑'
    font0.bold = False
    font0.colour_index = 0
    font0.height = 220
    style.font = font0
    # 设置文字位置
    alignment = xlwt.Alignment()  # 设置字体在单元格的位置
    alignment.horz = xlwt.Alignment.HORZ_CENTER  # 水平方向
    alignment.vert = xlwt.Alignment.VERT_CENTER  # 竖直方向
    style.alignment = alignment
    # 设置边框
    borders = xlwt.Borders()  # Create borders
    borders.left = xlwt.Borders.THIN  # 添加边框-虚线边框
    borders.right = xlwt.Borders.THIN  # 添加边框-虚线边框
    borders.top = xlwt.Borders.THIN  # 添加边框-虚线边框
    borders.bottom = xlwt.Borders.THIN  # 添加边框-虚线边框
    style.borders = borders

    return style
